__author__ = "Simon Nilsson"

import itertools
import os

import cv2
import numpy as np
import pandas as pd

import simba
from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.enums import Paths
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.warnings import ShapWarning


class ShapAggregateStatisticsVisualizer(ConfigReader):
    """
    Calculate aggregate (binned) SHAP value statistics where individual bins represent reaulated features.
    Also creates line chart visualizations reprsenting aggregations of behavior-present SHAP values.

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/SHAP.md#step-3-interpreting-the-shap-value-ouput-generated-by-simba>`__.
       `Example output <https://github.com/sgoldenlab/simba/blob/master/images/example_shap_graph.png>`__.
        .. image:: _static/img/shap.png
           :width: 400
           :align: center


    :parameter str config_path: Path to SimBA project config file in Configparser format
    :param str classifier_name: Name of classifier (e.g., Attack).
    :param pd.DataFrame shap_df: Dataframe with non-aggregated SHAP values where rows represent frames and columns represent features.
    :param float shap_baseline_value: SHAP expected value (computed by ``simba.train_model_functions.create_shap_log``).
    :param str save_path: Directory where to store the results




    :example:
    >>> _ = ShapAggregateStatisticsVisualizer(config_path='SimBAConfigFilePath', classifier_name='Attack', shap_df='tests/test_data/test_shap/data/test_shap.csv', shap_baseline_value=4, save_path='SaveDirectory')
    """

    def __init__(
        self,
        config_path: str,
        classifier_name: str,
        shap_df: pd.DataFrame,
        shap_baseline_value: int,
        save_path: str,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        if (self.pose_setting != "14") and (self.pose_setting != "16"):
            ShapWarning(
                "SHAP visualizations/aggregate stats skipped (only viable for projects with two animals and default 7 or 8 body-parts per animal) ..."
            )
        else:
            self.save_path = save_path
            self.classifier_name, self.shap_df, self.shap_baseline_value = (
                classifier_name,
                shap_df,
                shap_baseline_value,
            )
            if not os.path.exists(self.shap_logs_path):
                os.makedirs(self.shap_logs_path)
            self.img_save_path = os.path.join(
                self.shap_logs_path,
                "SHAP_summary_line_graph_{}_{}.png".format(
                    self.classifier_name, self.datetime
                ),
            )
            simba_dir = os.path.dirname(simba.__file__)
            feature_categories_csv_path = os.path.join(
                simba_dir, Paths.SIMBA_SHAP_CATEGORIES_PATH.value
            )
            shap_img_path = os.path.join(simba_dir, Paths.SIMBA_SHAP_IMG_PATH.value)
            check_file_exist_and_readable(file_path=feature_categories_csv_path)
            self.scale_img_dict = {
                "baseline_scale": os.path.join(shap_img_path, "baseline_scale.jpg"),
                "small_arrow": os.path.join(shap_img_path, "down_arrow.jpg"),
                "side_scale": os.path.join(shap_img_path, "side_scale.jpg"),
                "color_bar": os.path.join(shap_img_path, "color_bar.jpg"),
            }
            self.category_img_dict = {
                "Animal distances": {
                    "icon": os.path.join(shap_img_path, "animal_distances.jpg")
                },
                "Intruder movement": {
                    "icon": os.path.join(shap_img_path, "intruder_movement.jpg")
                },
                "Resident+intruder movement": {
                    "icon": os.path.join(
                        shap_img_path, "resident_intruder_movement.jpg"
                    )
                },
                "Resident movement": {
                    "icon": os.path.join(shap_img_path, "resident_movement.jpg")
                },
                "Intruder shape": {
                    "icon": os.path.join(shap_img_path, "intruder_shape.jpg")
                },
                "Resident+intruder shape": {
                    "icon": os.path.join(shap_img_path, "resident_intruder_shape.jpg")
                },
                "Resident shape": {
                    "icon": os.path.join(shap_img_path, "resident_shape.jpg")
                },
            }
            for k, v in self.scale_img_dict.items():
                check_file_exist_and_readable(file_path=v)
            for k, v in self.category_img_dict.items():
                check_file_exist_and_readable(file_path=v["icon"])
            self.positive_arrow_colors = [
                (253, 141, 60),
                (252, 78, 42),
                (227, 26, 28),
                (189, 0, 38),
                (128, 0, 38),
            ]
            self.negative_arrow_colors = [
                (65, 182, 196),
                (29, 145, 192),
                (34, 94, 168),
                (37, 52, 148),
                (8, 29, 88),
            ]
            self.ranges_lst = [
                list(range(0, 20)),
                list(range(20, 40)),
                list(range(40, 60)),
                list(range(60, 80)),
                list(range(80, 101)),
            ]
            self.feature_categories_df = pd.read_csv(
                feature_categories_csv_path, header=[0, 1]
            )
            self.unique_feature_category_names, self.unique_time_bin_names = set(
                list(self.feature_categories_df.columns.levels[0])
            ), set(list(self.feature_categories_df.columns.levels[1]))
            self.__calculate_agg_shap_scores()
            self.__create_base_shap_img()
            self.__insert_data_in_img()

    def __save_aggregate_scores(self):
        """
        Private helper to convert results in dict format to CSV file.
        """
        results = pd.DataFrame(
            columns=self.results[list(self.results.keys())[0]],
            index=self.results.keys(),
        )
        for row_name, time_bins in self.results.items():
            for column_name, value in time_bins.items():
                results.loc[row_name, column_name] = value
        results.reindex(sorted(results.columns, reverse=True), axis=1).to_csv(
            self.df_save_path
        )

    def __calculate_agg_shap_scores(self):
        """
        Private helper to aggregate SHAP values into bins of related features.
        """
        self.agg_stats_timer = SimbaTimer(start=True)
        for clf_state, clf_state_name in zip(range(2), ["ABSENT", "PRESENT"]):
            self.results = {}
            self.df_save_path = os.path.join(
                self.shap_logs_path,
                "SHAP_summary_{}_{}_{}.csv".format(
                    self.classifier_name, clf_state_name, self.datetime
                ),
            )
            shap_clf_sliced = self.shap_df[
                self.shap_df[self.classifier_name] == clf_state
            ]
            for feature_category, feature_time_bin in itertools.product(
                self.unique_feature_category_names, self.unique_time_bin_names
            ):
                if feature_category not in self.results.keys():
                    self.results[feature_category] = {}
                feature_names_sliced = list(
                    self.feature_categories_df.loc[
                        :, (feature_category, feature_time_bin)
                    ]
                )
                feature_names_sliced = [
                    x
                    for x in feature_names_sliced
                    if str(x) != "nan" and x in shap_clf_sliced
                ]
                self.results[feature_category][feature_time_bin] = round(
                    shap_clf_sliced[feature_names_sliced].sum(axis=1).mean() * 100, 6
                )
            self.__save_aggregate_scores()
        self.agg_stats_timer.stop_timer()
        self.visualization_timer = SimbaTimer(start=True)

        stdout_success(
            msg=f"Aggregate SHAP statistics saved in {self.shap_logs_path} directory",
            elapsed_time=self.agg_stats_timer.elapsed_time_str,
        )

    def __create_base_shap_img(self):
        """
        Private helper to create the base (axes, icons ticks etc) of the aggregate shap value visualization.
        """

        self.img = 255 * np.ones([1680, 1680, 3], dtype=np.uint8)
        self.baseline_scale_img = cv2.imread(self.scale_img_dict["baseline_scale"])
        self.baseline_scale_top_left = (100, 800)
        baseline_scale_bottom_right = (
            self.baseline_scale_top_left[0] + self.baseline_scale_img.shape[0],
            self.baseline_scale_top_left[1] + self.baseline_scale_img.shape[1],
        )
        baseline_scale_middle = ((int(700 + self.baseline_scale_img.shape[1] / 2)), 85)
        self.img[
            self.baseline_scale_top_left[0] : baseline_scale_bottom_right[0],
            self.baseline_scale_top_left[1] : baseline_scale_bottom_right[1],
        ] = self.baseline_scale_img
        cv2.putText(
            self.img,
            "BASELINE SHAP",
            baseline_scale_middle,
            cv2.FONT_HERSHEY_COMPLEX,
            1,
            (0, 0, 0),
            2,
        )

        self.small_arrow_img = cv2.imread(self.scale_img_dict["small_arrow"])
        small_arrow_top_left = (
            baseline_scale_bottom_right[0],
            int(
                self.baseline_scale_top_left[1]
                + (self.baseline_scale_img.shape[1] / 100) * (self.shap_baseline_value)
            ),
        )
        small_arrow_bottom_right = (
            small_arrow_top_left[0] + self.small_arrow_img.shape[0],
            small_arrow_top_left[1] + self.small_arrow_img.shape[1],
        )
        self.img[
            small_arrow_top_left[0] : small_arrow_bottom_right[0],
            small_arrow_top_left[1] : small_arrow_bottom_right[1],
        ] = self.small_arrow_img

        self.color_bar_img = cv2.imread(self.scale_img_dict["color_bar"])

        side_scale_img = cv2.imread(self.scale_img_dict["side_scale"])
        side_scale_top_left = (
            small_arrow_bottom_right[0] + 50,
            self.baseline_scale_top_left[1] - 50,
        )
        self.side_scale_y_tick_cords = [
            (side_scale_top_left[0], side_scale_top_left[1] - 75)
        ]
        for i in range(1, 7):
            self.side_scale_y_tick_cords.append(
                (
                    int(side_scale_top_left[0] + (side_scale_img.shape[0] / 4) * i),
                    int(side_scale_top_left[1] - 75),
                )
            )
        self.arrow_start = (
            int(small_arrow_top_left[1] + (self.small_arrow_img.shape[1] / 2)),
            side_scale_top_left[0],
        )

        for img_cnt, (img_name, img_data) in enumerate(self.category_img_dict.items()):
            icon_img = cv2.resize(
                cv2.imread(img_data["icon"]),
                None,
                fx=1.5,
                fy=1.5,
                interpolation=cv2.INTER_CUBIC,
            )
            icon_top_left = (
                self.side_scale_y_tick_cords[img_cnt][0] - int(icon_img.shape[0] / 2),
                self.side_scale_y_tick_cords[img_cnt][1] - 100,
            )
            icon_bottom_right = (
                icon_top_left[0] + icon_img.shape[0],
                self.side_scale_y_tick_cords[img_cnt][1] + icon_img.shape[1] - 100,
            )
            text_location = (
                int(
                    icon_bottom_right[0]
                    - (icon_bottom_right[0] - icon_top_left[0])
                    + 100
                ),
                int(icon_bottom_right[1] - (icon_bottom_right[1] - icon_top_left[1]))
                - 380,
            )
            cv2.putText(
                self.img,
                str(img_name),
                (text_location[1], text_location[0]),
                cv2.FONT_HERSHEY_COMPLEX,
                0.75,
                (0, 0, 0),
                1,
            )
            self.img[
                icon_top_left[0] : icon_bottom_right[0],
                icon_top_left[1] : icon_bottom_right[1],
            ] = icon_img

    def __insert_data_in_img(self):
        """
        Private helper to insert the data (i.e., colored arrows, text etc.) into the aggregate shap value visualization
        and save the results.
        """
        data_df = pd.read_csv(
            os.path.join(
                self.shap_logs_path,
                "SHAP_summary_{}_{}_{}.csv".format(
                    self.classifier_name, "PRESENT", self.datetime
                ),
            ),
            index_col=0,
        )
        for feature_category in self.unique_feature_category_names:
            self.category_img_dict[feature_category]["value"] = int(
                data_df.loc[feature_category, :].sum()
            )

        for row_cnt, (feature_category_name, feature_data) in enumerate(
            self.category_img_dict.items()
        ):
            arrow_width = int(
                (self.baseline_scale_img.shape[1] / 100) * abs(feature_data["value"])
            )
            if feature_data["value"] > 0:
                arrow_end = (self.arrow_start[0] + arrow_width, self.arrow_start[1])
                arrow_middle = int(
                    ((arrow_end[1] - self.arrow_start[1]) / 2) + self.arrow_start[1] - 7
                )
                for bracket_no, bracket in enumerate(self.ranges_lst):
                    if abs(feature_data["value"]) in bracket:
                        color = (
                            self.positive_arrow_colors[bracket_no][2],
                            self.positive_arrow_colors[bracket_no][1],
                            self.positive_arrow_colors[bracket_no][0],
                        )
                cv2.arrowedLine(
                    self.img, self.arrow_start, arrow_end, color, 5, tipLength=0.1
                )
                cv2.putText(
                    self.img,
                    "+" + str(abs(feature_data["value"])) + "%",
                    (arrow_end[0] - 7, arrow_middle - 15),
                    cv2.FONT_HERSHEY_COMPLEX,
                    1,
                    color,
                    2,
                )

            else:
                arrow_end = (self.arrow_start[0] - arrow_width, self.arrow_start[1])
                arrow_middle = int(
                    ((self.arrow_start[1] - arrow_end[1]) / 2) + arrow_end[1] - 7
                )
                for bracket_no, bracket in enumerate(self.ranges_lst):
                    if abs(feature_data["value"]) in bracket:
                        color = (
                            self.negative_arrow_colors[bracket_no][2],
                            self.negative_arrow_colors[bracket_no][1],
                            self.negative_arrow_colors[bracket_no][0],
                        )
                cv2.arrowedLine(
                    self.img, self.arrow_start, arrow_end, color, 5, tipLength=0.1
                )
                cv2.putText(
                    self.img,
                    "-" + str(abs(feature_data["value"])) + "%",
                    (arrow_end[0] - 7, arrow_middle - 15),
                    cv2.FONT_HERSHEY_COMPLEX,
                    1,
                    color,
                    2,
                )

            if row_cnt != (len(list(self.category_img_dict.keys())) - 1):
                self.arrow_start = (
                    arrow_end[0],
                    self.side_scale_y_tick_cords[row_cnt + 1][0],
                )

        small_arrow_top_left = (
            int(arrow_end[1]) + 20,
            int(arrow_end[0] - self.small_arrow_img.shape[1] / 2),
        )
        small_arrow_bottom_right = (
            small_arrow_top_left[0] + self.small_arrow_img.shape[0],
            small_arrow_top_left[1] + self.small_arrow_img.shape[1],
        )
        self.img[
            small_arrow_top_left[0] : small_arrow_bottom_right[0],
            small_arrow_top_left[1] : small_arrow_bottom_right[1],
        ] = self.small_arrow_img
        color_bar_top_left = (
            arrow_end[1] + self.small_arrow_img.shape[0] + 25,
            self.baseline_scale_top_left[1],
        )
        color_bar_bottom_right = (
            color_bar_top_left[0] + self.color_bar_img.shape[0],
            color_bar_top_left[1] + self.color_bar_img.shape[1],
        )
        self.img[
            color_bar_top_left[0] : color_bar_bottom_right[0],
            color_bar_top_left[1] : color_bar_bottom_right[1],
        ] = self.color_bar_img

        color_bar_middle = (
            (int(580 + self.baseline_scale_img.shape[1] / 2)),
            color_bar_bottom_right[0] + 50,
        )
        cv2.putText(
            self.img,
            "CLASSIFICATION PROBABILITY",
            color_bar_middle,
            cv2.FONT_HERSHEY_COMPLEX,
            1,
            (0, 0, 0),
            2,
        )
        cv2.imwrite(self.img_save_path, self.img)
        self.visualization_timer.stop_timer()
        stdout_success(
            msg=f"SHAP summary graph saved at {self.img_save_path}",
            elapsed_time=self.visualization_timer.elapsed_time_str,
        )


# shap_df = pd.read_csv('/Users/simon/Desktop/envs/simba_dev/tests/test_data/test_shap/data/test_shap.csv', index_col=0)
# test = ShapAggregateStatisticsVisualizer(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                          classifier_name='Attack',
#                                          shap_df=shap_df,
#                                          shap_baseline_value=40,
#                                          save_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/logs/shap')



__author__ = "Simon Nilsson"


import os
from itertools import product

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import check_str
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df


class UserDefinedFeatureExtractor(ConfigReader, FeatureExtractionMixin):
    """
    Generic featurizer of data within SimBA project using user-defined body-parts in the pose-estimation data.
    Results are stored in the `project_folder/csv/features_extracted` directory of the SimBA project.

    :parameter str config_path: path to SimBA project config file in Configparser format

    .. note::
       `Feature extraction tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-5-extract-features>`__.

    Examples
    ----------
    >>> feature_extractor = UserDefinedFeatureExtractor(config_path='MyProjectConfig')
    >>> feature_extractor.run()
    """

    def __init__(self, config_path: str):
        FeatureExtractionMixin.__init__(self, config_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        print(
            "Extracting features from {} file(s)...".format(str(len(self.files_found)))
        )

    def __euclid_dist_between_bps_of_other_animals(self):
        print("Calculating euclidean distances...")
        self.distance_col_names = []
        for animal_name, animal_data in self.animal_bp_dict.items():
            current_animal_bp_xs, current_animal_bp_ys = (
                animal_data["X_bps"],
                animal_data["Y_bps"],
            )
            other_animals = {
                i: self.animal_bp_dict[i]
                for i in self.animal_bp_dict
                if i != animal_name
            }
            for current_animal_bp_x, current_animal_bp_y in zip(
                current_animal_bp_xs, current_animal_bp_ys
            ):
                for other_animal_name, other_animal_data in other_animals.items():
                    other_animal_bp_xs, other_animal_bp_ys = (
                        other_animal_data["X_bps"],
                        other_animal_data["Y_bps"],
                    )
                    for other_animal_bp_x, other_animal_bp_y in zip(
                        other_animal_bp_xs, other_animal_bp_ys
                    ):
                        current_bp_name, other_bp_name = current_animal_bp_x.strip(
                            "_x"
                        ), other_animal_bp_x.strip("_x")
                        col_name = "Euclidean_distance_{}_{}".format(
                            current_bp_name, other_bp_name
                        )
                        reverse_col_name = "Euclidean_distance_{}_{}".format(
                            other_bp_name, current_bp_name
                        )
                        if not reverse_col_name in self.data_df.columns:
                            self.data_df[col_name] = (
                                np.sqrt(
                                    (
                                        self.data_df[current_animal_bp_x]
                                        - self.data_df[other_animal_bp_x]
                                    )
                                    ** 2
                                    + (
                                        self.data_df[current_animal_bp_y]
                                        - self.data_df[other_animal_bp_y]
                                    )
                                    ** 2
                                )
                            ) / self.px_per_mm
                            self.distance_col_names.append(col_name)

    def __movement_of_all_bps(self):
        print("Calculating movements of all body-parts...")
        self.mean_movement_cols, self.sum_movement_cols = [], []
        for animal_name, animal_data in self.animal_bp_dict.items():
            animal_cols = []
            current_animal_bp_xs, current_animal_bp_ys = (
                animal_data["X_bps"],
                animal_data["Y_bps"],
            )
            for bp_x, bp_y in zip(current_animal_bp_xs, current_animal_bp_ys):
                shifted_bp_x, shifted_bp_y = bp_x + "_shifted", bp_y + "_shifted"
                col_name = "Movement_" + bp_x.strip("_x")
                self.data_df[col_name] = (
                    np.sqrt(
                        (self.data_df_comb[bp_x] - self.data_df_comb[shifted_bp_x]) ** 2
                        + (self.data_df_comb[bp_y] - self.data_df_comb[shifted_bp_y])
                        ** 2
                    )
                ) / self.px_per_mm
                animal_cols.append(col_name)
            self.data_df["All_bp_movements_" + animal_name + "_sum"] = self.data_df[
                animal_cols
            ].sum(axis=1)
            self.data_df["All_bp_movements_" + animal_name + "_mean"] = self.data_df[
                animal_cols
            ].mean(axis=1)
            self.data_df["All_bp_movements_" + animal_name + "_min"] = self.data_df[
                animal_cols
            ].min(axis=1)
            self.data_df["All_bp_movements_" + animal_name + "_max"] = self.data_df[
                animal_cols
            ].max(axis=1)
            self.mean_movement_cols.append("All_bp_movements_" + animal_name + "_mean")
            self.sum_movement_cols.append("All_bp_movements_" + animal_name + "_sum")

    def __rolling_windows_bp_distances(self):
        print("Calculating rolling windows data: distances between body-parts...")
        for i in product(self.roll_windows_values, self.distance_col_names):
            self.data_df["Mean_{}_{}".format(i[1], i[0])] = (
                self.data_df[i[1]].rolling(int(i[0]), min_periods=1).mean()
            )
            self.data_df["Sum_{}_{}".format(i[1], i[0])] = (
                self.data_df[i[1]].rolling(int(i[0]), min_periods=1).sum()
            )

    def __rolling_windows_movement(self):
        print("Calculating rolling windows data: animal movements...")
        for i in product(self.roll_windows_values, self.mean_movement_cols):
            self.data_df["Mean_{}_{}".format(i[1], i[0])] = (
                self.data_df[i[1]].rolling(int(i[0]), min_periods=1).mean()
            )
            self.data_df["Sum_{}_{}".format(i[1], i[0])] = (
                self.data_df[i[1]].rolling(int(i[0]), min_periods=1).sum()
            )

    def __pose_probability_filters(self):
        p_df = self.data_df.filter(self.p_cols, axis=1)
        self.data_df["Sum_probabilities"] = p_df.sum(axis=1)
        self.data_df["Mean_probabilities"] = p_df.mean(axis=1)
        results = pd.DataFrame(
            self.count_values_in_range(
                data=self.data_df.filter(self.p_cols).values,
                ranges=np.array([[0.0, 0.1], [0.0, 0.5], [0.0, 0.75]]),
            ),
            columns=[
                "Low_prob_detections_0.1",
                "Low_prob_detections_0.5",
                "Low_prob_detections_0.75",
            ],
        )
        self.data_df = pd.concat([self.data_df, results], axis=1)

    def run(self):
        """
        Method to compute and save features to disk. Results are saved in the `project_folder/csv/features_extracted`
        directory of the SimBA project.

        Returns
        -------
        None
        """

        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            print(
                "Extracting features for video {}/{}...".format(
                    str(file_cnt + 1), str(len(self.files_found))
                )
            )
            _, file_name, _ = get_fn_ext(file_path)
            check_str("file name", file_name)
            video_settings, self.px_per_mm, fps = self.read_video_info(
                video_name=file_name
            )
            roll_windows = []
            for i in range(len(self.roll_windows_values)):
                roll_windows.append(int(fps / self.roll_windows_values[i]))
            self.data_df = read_df(file_path, self.file_type)
            self.data_df.columns = self.col_headers
            self.data_df = self.data_df.fillna(0).apply(pd.to_numeric)
            self.data_df_shifted = self.data_df.shift(periods=1)
            self.data_df_shifted.columns = self.col_headers_shifted
            self.data_df_comb = (
                pd.concat([self.data_df, self.data_df_shifted], axis=1, join="inner")
                .fillna(0)
                .reset_index(drop=True)
            )
            self.__euclid_dist_between_bps_of_other_animals()
            self.__movement_of_all_bps()
            self.__rolling_windows_bp_distances()
            self.__rolling_windows_movement()
            self.__pose_probability_filters()
            save_path = os.path.join(self.save_dir, file_name + "." + self.file_type)
            self.data_df = self.data_df.reset_index(drop=True).fillna(0)
            write_df(df=self.data_df, file_type=self.file_type, save_path=save_path)
            video_timer.stop_timer()
            print(
                f"Feature extraction complete for video {file_name} (elapsed time: {video_timer.elapsed_time_str}s)"
            )

        self.timer.stop_timer()
        stdout_success(
            f"Feature extraction complete for {str(len(self.files_found))} video(s). Results are saved inside the project_folder/csv/features_extracted directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = UserDefinedFeatureExtractor(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.run()



__author__ = "Simon Nilsson"

import os
from collections import defaultdict
from tkinter import *

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.distance_plotter import DistancePlotterSingleCore
from simba.plotting.distance_plotter_mp import DistancePlotterMultiCore
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box)
from simba.utils.checks import check_if_filepath_list_is_empty, check_int
from simba.utils.enums import Formats, Keys, Links
from simba.utils.errors import DuplicationError
from simba.utils.read_write import get_file_name_info_in_directory


class DistancePlotterPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="CREATE DISTANCE PLOTS")
        ConfigReader.__init__(self, config_path=config_path)

        self.data_path = os.path.join(
            self.project_path, "csv", "outlier_corrected_movement_location"
        )
        self.max_y_lst = list(range(10, 510, 10))
        self.max_y_lst.insert(0, "auto")
        self.files_found_dict = get_file_name_info_in_directory(
            directory=self.data_path, file_type=self.file_type
        )
        check_if_filepath_list_is_empty(
            filepaths=list(self.files_found_dict.keys()),
            error_msg="SIMBA ERROR: Zero files found in the project_folder/csv/outlier_corrected_movement_location directory. ",
        )

        self.number_of_distances = list(range(1, len(self.body_parts_lst) * 2))
        self.style_settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="STYLE SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.DISTANCE_PLOTS.value,
        )
        self.resolution_dropdown = DropDownMenu(
            self.style_settings_frm, "Resolution:", self.resolutions, "16"
        )
        self.font_size_entry = Entry_Box(
            self.style_settings_frm, "Font size: ", "16", validation="numeric"
        )
        self.line_width = Entry_Box(
            self.style_settings_frm, "Line width: ", "16", validation="numeric"
        )
        self.opacity_dropdown = DropDownMenu(
            self.style_settings_frm,
            "Line opacity:",
            list(np.round(np.arange(0.0, 1.1, 0.1), 1)),
            "16",
        )
        self.max_y_dropdown = DropDownMenu(
            self.style_settings_frm, "Max Y-axis:", self.max_y_lst, "16"
        )
        self.resolution_dropdown.setChoices(self.resolutions[1])
        self.font_size_entry.entry_set(val=8)
        self.line_width.entry_set(val=6)
        self.opacity_dropdown.setChoices(0.5)
        self.max_y_dropdown.setChoices(choice="auto")
        self.distances_frm = LabelFrame(
            self.main_frm,
            text="CHOOSE DISTANCES",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
        )
        self.number_of_distances_dropdown = DropDownMenu(
            self.distances_frm,
            "# Distances:",
            self.number_of_distances,
            "16",
            com=self.__populate_distances_menu,
        )
        self.number_of_distances_dropdown.setChoices(self.number_of_distances[0])
        self.__populate_distances_menu(1)

        self.settings_frm = LabelFrame(
            self.main_frm,
            text="VISUALIZATION SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
        )
        self.distance_frames_var = BooleanVar()
        self.distance_videos_var = BooleanVar()
        self.distance_final_img_var = BooleanVar()
        self.multiprocess_var = BooleanVar()
        distance_frames_cb = Checkbutton(
            self.settings_frm, text="Create frames", variable=self.distance_frames_var
        )
        distance_videos_cb = Checkbutton(
            self.settings_frm, text="Create videos", variable=self.distance_videos_var
        )
        distance_final_img_cb = Checkbutton(
            self.settings_frm,
            text="Create last frame",
            variable=self.distance_final_img_var,
        )
        self.multiprocess_cb = Checkbutton(
            self.settings_frm,
            text="Multiprocess (faster)",
            variable=self.multiprocess_var,
            command=lambda: self.enable_dropdown_from_checkbox(
                check_box_var=self.multiprocess_var,
                dropdown_menus=[self.multiprocess_dropdown],
            ),
        )
        self.multiprocess_dropdown = DropDownMenu(
            self.settings_frm, "Cores:", list(range(2, self.cpu_cnt)), "12"
        )
        self.multiprocess_dropdown.setChoices(choice=2)
        self.multiprocess_dropdown.disable()

        self.run_frm = LabelFrame(
            self.main_frm,
            text="RUN",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_single_video_frm = LabelFrame(
            self.run_frm,
            text="SINGLE VIDEO",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_single_video_btn = Button(
            self.run_single_video_frm,
            text="Create single video",
            fg="blue",
            command=lambda: self.__create_distance_plots(multiple_videos=False),
        )
        self.single_video_dropdown = DropDownMenu(
            self.run_single_video_frm,
            "Video:",
            list(self.files_found_dict.keys()),
            "12",
        )
        self.single_video_dropdown.setChoices(list(self.files_found_dict.keys())[0])

        self.run_multiple_videos = LabelFrame(
            self.run_frm,
            text="MULTIPLE VIDEO",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_multiple_video_btn = Button(
            self.run_multiple_videos,
            text="Create multiple videos ({} video(s) found)".format(
                str(len(list(self.files_found_dict.keys())))
            ),
            fg="blue",
            command=lambda: self.__create_distance_plots(multiple_videos=True),
        )

        self.style_settings_frm.grid(row=0, sticky=NW)
        self.resolution_dropdown.grid(row=0, sticky=NW)
        self.font_size_entry.grid(row=1, sticky=NW)
        self.line_width.grid(row=2, sticky=NW)
        self.opacity_dropdown.grid(row=3, sticky=NW)
        self.max_y_dropdown.grid(row=4, sticky=NW)

        self.distances_frm.grid(row=1, sticky=NW)
        self.number_of_distances_dropdown.grid(row=0, sticky=NW)

        self.settings_frm.grid(row=2, sticky=NW)
        distance_frames_cb.grid(row=0, sticky=NW)
        distance_videos_cb.grid(row=1, sticky=NW)
        distance_final_img_cb.grid(row=2, sticky=NW)
        self.multiprocess_cb.grid(row=3, column=0, sticky=NW)
        self.multiprocess_dropdown.grid(row=3, column=1, sticky=NW)

        self.run_frm.grid(row=3, sticky=NW)
        self.run_single_video_frm.grid(row=0, sticky=NW)
        self.run_single_video_btn.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)

        self.run_multiple_videos.grid(row=1, sticky=NW)
        self.run_multiple_video_btn.grid(row=0, sticky=NW)

        self.main_frm.mainloop()

    def __populate_distances_menu(self, choice):
        if hasattr(self, "bp_1"):
            for k, v in self.bp_1.items():
                self.bp_1[k].destroy()
                self.bp_2[k].destroy()
                self.distance_clrs[k].destroy()

        self.bp_1, self.bp_2, self.distance_clrs = {}, {}, {}
        for distance_cnt in range(int(self.number_of_distances_dropdown.getChoices())):
            self.bp_1[distance_cnt] = DropDownMenu(
                self.distances_frm,
                "Distance {}:".format(str(distance_cnt + 1)),
                self.body_parts_lst,
                "16",
            )
            self.bp_1[distance_cnt].setChoices(self.body_parts_lst[distance_cnt])
            self.bp_1[distance_cnt].grid(row=distance_cnt + 1, column=0, sticky=NW)

            self.bp_2[distance_cnt] = DropDownMenu(
                self.distances_frm, "", self.body_parts_lst, "2"
            )
            self.bp_2[distance_cnt].setChoices(self.body_parts_lst[distance_cnt])
            self.bp_2[distance_cnt].grid(row=distance_cnt + 1, column=1, sticky=NW)

            self.distance_clrs[distance_cnt] = DropDownMenu(
                self.distances_frm, "", self.colors_dict, "2"
            )
            self.distance_clrs[distance_cnt].setChoices(
                list(self.colors_dict.keys())[distance_cnt]
            )
            self.distance_clrs[distance_cnt].grid(
                row=distance_cnt + 1, column=3, sticky=NW
            )

    def __create_distance_plots(self, multiple_videos: bool):
        if multiple_videos:
            data_paths = list(self.files_found_dict.values())
        else:
            data_paths = [
                self.files_found_dict[self.single_video_dropdown.getChoices()]
            ]

        line_attr = defaultdict(list)
        for attr in (self.bp_1, self.bp_2, self.distance_clrs):
            for key, value in attr.items():
                line_attr[key].append(value.getChoices())

        for cnt, (k, v) in enumerate(line_attr.items()):
            if v[0] == v[1]:
                raise DuplicationError(
                    msg=f"DISTANCE LINE {cnt+1} ERROR: The two body-parts cannot be the same body-part.",
                    source=self.__class__.__name__,
                )

        width = int(self.resolution_dropdown.getChoices().split("×")[0])
        height = int(self.resolution_dropdown.getChoices().split("×")[1])
        check_int(
            name="DISTANCE FONT SIZE", value=self.font_size_entry.entry_get, min_value=1
        )
        check_int(
            name="DISTANCE LINE WIDTH", value=self.line_width.entry_get, min_value=1
        )
        style_attr = {
            "width": width,
            "height": height,
            "line width": int(self.line_width.entry_get),
            "font size": int(self.font_size_entry.entry_get),
            "opacity": float(self.opacity_dropdown.getChoices()),
            "y_max": self.max_y_dropdown.getChoices(),
        }
        if not self.multiprocess_var.get():
            distance_plotter = DistancePlotterSingleCore(
                config_path=self.config_path,
                frame_setting=self.distance_frames_var.get(),
                video_setting=self.distance_videos_var.get(),
                final_img=self.distance_final_img_var.get(),
                style_attr=style_attr,
                files_found=data_paths,
                line_attr=line_attr,
            )
        else:
            distance_plotter = DistancePlotterMultiCore(
                config_path=self.config_path,
                frame_setting=self.distance_frames_var.get(),
                video_setting=self.distance_videos_var.get(),
                final_img=self.distance_final_img_var.get(),
                style_attr=style_attr,
                files_found=data_paths,
                line_attr=line_attr,
                core_cnt=int(self.multiprocess_dropdown.getChoices()),
            )

        distance_plotter.run()


# _ = DistancePlotterPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# _ = DistancePlotterPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import os.path
import platform
import webbrowser
from tkinter import *
from tkinter.filedialog import askdirectory, askopenfilename
from typing import Optional, Union

import PIL.Image
from PIL import ImageTk

from simba.utils.enums import Defaults, Formats
from simba.utils.lookups import get_icons_paths

MENU_ICONS = get_icons_paths()


def onMousewheel(event, canvas):
    try:
        scrollSpeed = event.delta
        if platform.system() == "Darwin":
            scrollSpeed = event.delta
        elif platform.system() == "Windows":
            scrollSpeed = int(event.delta / 120)
        canvas.yview_scroll(-1 * (scrollSpeed), "units")
    except:
        pass


def bindToMousewheel(event, canvas):
    canvas.bind_all("<MouseWheel>", lambda event: onMousewheel(event, canvas))


def unbindToMousewheel(event, canvas):
    canvas.unbind_all("<MouseWheel>")


def onFrameConfigure(canvas):
    """Reset the scroll region to encompass the inner frame"""
    canvas.configure(scrollregion=canvas.bbox("all"))


def hxtScrollbar(master):
    """
    Create canvas.
    Create a frame and put it in the canvas.
    Create two scrollbar and insert command of canvas x and y view
    Use canvas to create a window, where window = frame
    Bind the frame to the canvas
    """
    bg = master.cget("background")
    acanvas = Canvas(
        master,
        borderwidth=0,
        background=bg,
        width=master.winfo_width(),
        height=master.winfo_reqheight(),
    )
    frame = Frame(acanvas, background=bg)
    vsb = Scrollbar(master, orient="vertical", command=acanvas.yview)
    vsb2 = Scrollbar(master, orient="horizontal", command=acanvas.xview)
    acanvas.configure(yscrollcommand=vsb.set)
    acanvas.configure(xscrollcommand=vsb2.set)
    vsb.pack(side=RIGHT, fill="y")
    vsb2.pack(side=BOTTOM, fill="x")
    acanvas.pack(side=LEFT, fill=BOTH, expand=True)

    acanvas.create_window((10, 10), window=frame, anchor="nw")

    # bind the frame to the canvas
    acanvas.bind("<Configure>", lambda event, canvas=acanvas: onFrameConfigure(acanvas))
    acanvas.bind("<Enter>", lambda event: bindToMousewheel(event, acanvas))
    acanvas.bind("<Leave>", lambda event: unbindToMousewheel(event, acanvas))
    acanvas.update()
    return frame


def form_validator_is_numeric(inStr, acttyp):
    if acttyp == "1":  # insert
        if not inStr.isdigit():
            return False
    return True


class DropDownMenu(Frame):
    def __init__(
        self,
        parent=None,
        dropdownLabel="",
        choice_dict=None,
        labelwidth="",
        com=None,
        **kw
    ):
        Frame.__init__(self, master=parent, **kw)
        self.dropdownvar = StringVar()
        self.lblName = Label(self, text=dropdownLabel, width=labelwidth, anchor=W)
        self.lblName.grid(row=0, column=0)
        self.choices = choice_dict
        self.popupMenu = OptionMenu(self, self.dropdownvar, *self.choices, command=com)
        self.popupMenu.grid(row=0, column=1)

    def getChoices(self):
        return self.dropdownvar.get()

    def setChoices(self, choice):
        self.dropdownvar.set(choice)

    def enable(self):
        self.popupMenu.configure(state="normal")

    def disable(self):
        self.popupMenu.configure(state="disable")


class FileSelect(Frame):
    def __init__(
        self,
        parent=None,
        fileDescription="",
        color=None,
        title=None,
        lblwidth=None,
        file_types=None,
        dropdown: DropDownMenu = None,
        initialdir: Optional[Union[str, os.PathLike]] = None,
        **kw
    ):

        self.title, self.dropdown, self.initialdir = title, dropdown, initialdir
        self.file_type = file_types
        self.color = color if color is not None else "black"
        self.lblwidth = lblwidth if lblwidth is not None else 0
        self.parent = parent
        Frame.__init__(self, master=parent, **kw)
        browse_icon = ImageTk.PhotoImage(
            image=PIL.Image.open(MENU_ICONS["browse"]["icon_path"])
        )
        self.filePath = StringVar()
        self.lblName = Label(
            self,
            text=fileDescription,
            fg=str(self.color),
            width=str(self.lblwidth),
            anchor=W,
        )
        self.lblName.grid(row=0, column=0, sticky=W)
        self.entPath = Label(self, textvariable=self.filePath, relief=SUNKEN)
        self.entPath.grid(row=0, column=1)
        self.btnFind = Button(
            self,
            text=Defaults.BROWSE_FILE_BTN_TEXT.value,
            compound="left",
            image=browse_icon,
            relief=RAISED,
            command=self.setFilePath,
        )
        self.btnFind.image = browse_icon
        self.btnFind.grid(row=0, column=2)
        self.filePath.set(Defaults.NO_FILE_SELECTED_TEXT.value)

    def setFilePath(self):
        if self.initialdir is not None:
            if not os.path.isdir(self.initialdir):
                self.initialdir = None
            else:
                pass

        if self.file_type:
            file_selected = askopenfilename(
                title=self.title,
                parent=self.parent,
                filetypes=self.file_type,
                initialdir=self.initialdir,
            )
        else:
            file_selected = askopenfilename(
                title=self.title, parent=self.parent, initialdir=self.initialdir
            )
        if file_selected:
            if self.dropdown is not None:
                self.dropdown.setChoices(os.path.basename(file_selected))
                self.filePath.set(os.path.basename(file_selected))
            else:
                self.filePath.set(file_selected)

        else:
            self.filePath.set(Defaults.NO_FILE_SELECTED_TEXT.value)

    @property
    def file_path(self):
        return self.filePath.get()

    def set_state(self, setstatus):
        self.entPath.config(state=setstatus)
        self.btnFind["state"] = setstatus


class Entry_Box(Frame):
    def __init__(
        self,
        parent=None,
        fileDescription="",
        labelwidth="",
        status=None,
        validation=None,
        entry_box_width=None,
        **kw
    ):
        super(Entry_Box, self).__init__(master=parent)
        self.validation_methods = {
            "numeric": (self.register(form_validator_is_numeric), "%P", "%d"),
        }
        self.status = status if status is not None else NORMAL
        self.labelname = fileDescription
        Frame.__init__(self, master=parent, **kw)
        self.filePath = StringVar()
        self.lblName = Label(self, text=fileDescription, width=labelwidth, anchor=W)
        self.lblName.grid(row=0, column=0)
        if not entry_box_width:
            self.entPath = Entry(
                self,
                textvariable=self.filePath,
                state=self.status,
                validate="key",
                validatecommand=self.validation_methods.get(validation, None),
            )
        else:
            self.entPath = Entry(
                self,
                textvariable=self.filePath,
                state=self.status,
                width=entry_box_width,
                validate="key",
                validatecommand=self.validation_methods.get(validation, None),
            )

        self.entPath.grid(row=0, column=1)

    @property
    def entry_get(self):
        self.entPath.get()
        return self.entPath.get()

    def entry_set(self, val):
        self.filePath.set(val)

    def set_state(self, setstatus):
        self.entPath.config(state=setstatus)

    def destroy(self):
        try:
            self.lblName.destroy()
            self.entPath.destroy()
        except:
            pass


class FolderSelect(Frame):
    def __init__(
        self,
        parent: Frame,
        folderDescription: Optional[str] = "",
        color: Optional[str] = None,
        title: Optional[str] = "",
        lblwidth: Optional[int] = 0,
        initialdir: Optional[Union[str, os.PathLike]] = None,
        **kw
    ):

        self.title, self.initialdir = title, initialdir
        self.color = color if color is not None else "black"
        self.lblwidth = lblwidth if lblwidth is not None else 0
        self.parent = parent
        Frame.__init__(self, master=parent, **kw)
        browse_icon = ImageTk.PhotoImage(
            image=PIL.Image.open(MENU_ICONS["browse"]["icon_path"])
        )
        self.folderPath = StringVar()
        self.lblName = Label(
            self,
            text=folderDescription,
            fg=str(self.color),
            width=str(self.lblwidth),
            anchor=W,
        )
        self.lblName.grid(row=0, column=0, sticky=W)
        self.entPath = Label(self, textvariable=self.folderPath, relief=SUNKEN)
        self.entPath.grid(row=0, column=1)
        self.btnFind = Button(
            self,
            text=Defaults.BROWSE_FOLDER_BTN_TEXT.value,
            compound="left",
            image=browse_icon,
            relief=RAISED,
            command=self.setFolderPath,
        )
        self.btnFind.image = browse_icon
        self.btnFind.grid(row=0, column=2)
        self.folderPath.set("No folder selected")

    def setFolderPath(self):
        if self.initialdir is not None:
            if not os.path.isdir(self.initialdir):
                self.initialdir = None
            else:
                pass
        folder_selected = askdirectory(
            title=str(self.title), parent=self.parent, initialdir=self.initialdir
        )
        if folder_selected:
            self.folderPath.set(folder_selected)
        else:
            self.folderPath.set("No folder selected")

    @property
    def folder_path(self):
        return self.folderPath.get()


class ToolTip(object):

    def __init__(self, widget):
        self.widget = widget
        self.tipwindow = None
        self.id = None
        self.x = self.y = 0

    def showtip(self, text):
        "Display text in tooltip window"
        self.text = text
        if self.tipwindow or not self.text:
            return
        x, y, cx, cy = self.widget.bbox("insert")
        x = x + self.widget.winfo_rootx() + 57
        y = y + cy + self.widget.winfo_rooty() + 27
        self.tipwindow = tw = Toplevel(self.widget)
        tw.wm_overrideredirect(1)
        tw.wm_geometry("+%d+%d" % (x, y))
        label = Label(
            tw,
            text=self.text,
            justify=LEFT,
            background="#ffffe0",
            relief=SOLID,
            borderwidth=1,
            font=("tahoma", "8", "normal"),
        )
        label.pack(ipadx=1)

    def hidetip(self):
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            tw.destroy()


def CreateToolTip(widget, text):
    toolTip = ToolTip(widget)

    def enter(event):
        toolTip.showtip(text)

    def leave(event):
        toolTip.hidetip()

    widget.bind("<Enter>", enter)
    widget.bind("<Leave>", leave)


def CreateLabelFrameWithIcon(
    parent: Toplevel, header: str, icon_name: str, icon_link: str or None = None
):

    icon = PIL.Image.open(MENU_ICONS[icon_name]["icon_path"])
    icon = ImageTk.PhotoImage(icon)
    frm = Frame(parent)
    label_text = Label(frm, text=header, font=Formats.LABELFRAME_HEADER_FORMAT.value)
    label_text.grid(row=0, column=0)
    label_image = Label(frm, image=icon)
    label_image.image = icon
    if icon_link:
        label_image.bind("<Button-1>", lambda e: callback(icon_link))
    label_image.grid(row=0, column=1)
    return LabelFrame(parent, labelwidget=frm)


def callback(url):
    webbrowser.open_new(url)


def create_scalebar(
    parent: Frame, name: str, min: int, max: int, cmd: object or None = None
):

    scale = Scale(
        parent,
        from_=min,
        to=max,
        orient=HORIZONTAL,
        length=200,
        label=name,
        command=cmd,
    )
    scale.set(0)
    return scale


class TwoOptionQuestionPopUp(object):
    """
    Helpe to create a two-option question tkinter pop up window (e.g., YES/NO).

    :parameter str question: Question to present to the user. E.g., ``DO YOU WANT TO PROCEED?``.
    :parameter str option_one: The first user option. E.g., ``YES``.
    :parameter str option_one: The second user option. E.g., ``NO``.
    :parameter str title: The window titel in the top banner of the pop-up. E.g., ``A QUESTION FOR YOU!``.
    :parameter Optional[str] link: If not None, then a link to documentation presenting background info about the user choices.
    """

    def __init__(
        self,
        question: str,
        option_one: str,
        option_two: str,
        title: str,
        link: Optional[str] = None,
    ):

        self.main_frm = Toplevel()
        self.main_frm.geometry("600x200")
        self.main_frm.title(title)

        question_frm = Frame(self.main_frm)
        question_frm.pack(expand=True, fill="both")
        Label(
            question_frm, text=question, font=Formats.LABELFRAME_HEADER_FORMAT.value
        ).pack()
        button_one = Button(
            question_frm,
            text=option_one,
            fg="blue",
            command=lambda: self.run(option_one),
        )
        button_two = Button(
            question_frm,
            text=option_two,
            fg="red",
            command=lambda: self.run(option_two),
        )
        if link:
            link_lbl = Label(
                question_frm,
                text="Click here for more information.",
                cursor="hand2",
                fg="blue",
            )
            link_lbl.bind("<Button-1>", lambda e: callback(link))
            link_lbl.place(relx=0.5, rely=0.30, anchor=CENTER)
        button_one.place(relx=0.5, rely=0.50, anchor=CENTER)
        button_two.place(relx=0.5, rely=0.70, anchor=CENTER)
        self.main_frm.wait_window()

    def run(self, selected_option):
        self.selected_option = selected_option
        self.main_frm.destroy()



__author__ = "Simon Nilsson"

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import logging
import time
from typing import Optional

from simba.utils.enums import Defaults, TagNames


def stdout_success(
    msg: str, source: str = "", elapsed_time: Optional[str] = None
) -> None:
    log_event(
        logger_name=f"{source}.{stdout_success.__name__}",
        log_type=TagNames.COMPLETE.value,
        msg=msg,
    )
    if elapsed_time:
        print(
            f"SIMBA COMPLETE: {msg} (elapsed time: {elapsed_time}s) {Defaults.STR_SPLIT_DELIMITER.value}{TagNames.COMPLETE.value}"
        )
    else:
        print(
            f"SIMBA COMPLETE: {msg} {Defaults.STR_SPLIT_DELIMITER.value}{TagNames.COMPLETE.value}"
        )


def stdout_warning(msg: str, elapsed_time: Optional[str] = None) -> None:
    if elapsed_time:
        print(
            f"SIMBA WARNING: {msg} (elapsed time: {elapsed_time}s) {Defaults.STR_SPLIT_DELIMITER.value}{TagNames.WARNING.value}"
        )
    else:
        print(
            f"SIMBA WARNING: {msg} {Defaults.STR_SPLIT_DELIMITER.value}{TagNames.WARNING.value}"
        )


def stdout_trash(
    msg: str, source: str = "", elapsed_time: Optional[str] = None
) -> None:
    log_event(
        logger_name=f"{source}.{stdout_trash.__name__}",
        log_type=TagNames.TRASH.value,
        msg=msg,
    )
    if elapsed_time:
        print(
            f"SIMBA COMPLETE: {msg} (elapsed time: {elapsed_time}s) {Defaults.STR_SPLIT_DELIMITER.value}{TagNames.TRASH.value}"
        )
    else:
        print(
            f"SIMBA COMPLETE: {msg} {Defaults.STR_SPLIT_DELIMITER.value}{TagNames.TRASH.value}"
        )


class SimbaTimer(object):
    def __init__(self, start: bool = False):
        if start:
            self.start_timer()

    def start_timer(self):
        self.timer = time.time()

    def stop_timer(self):
        if not hasattr(self, "timer"):
            self.elapsed_time = -1
            self.elapsed_time_str = "-1"
        else:
            self.elapsed_time = round(time.time() - self.timer, 4)
            self.elapsed_time_str = str(self.elapsed_time)


def log_event(
    logger_name: str, log_type: Literal["CLASS_INIT", "error", "warning"], msg: str
):
    logger = logging.getLogger(str(logger_name))
    if log_type == TagNames.CLASS_INIT.value:
        logger.info(f"{TagNames.CLASS_INIT.value}||{msg}")
    elif log_type == TagNames.ERROR.value:
        logger.error(f"{TagNames.ERROR.value}||{msg}")
    elif log_type == TagNames.WARNING.value:
        logger.warning(f"{TagNames.WARNING.value}||{msg}")
    elif log_type == TagNames.TRASH.value:
        logger.info(f"{TagNames.TRASH.value}||{msg}")
    elif log_type == TagNames.COMPLETE.value:
        logger.info(f"{TagNames.COMPLETE.value}||{msg}")


def perform_timing(func):
    def decorator(*args, **kwargs):
        timer = SimbaTimer(start=True)
        results = func(*args, **kwargs, _timer=timer)
        timer.stop_timer()
        results["timer"] = timer.elapsed_time_str
        return results

    return decorator



import ast
import os
from itertools import product
from typing import Dict, Union

from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from simba.mixins.config_reader import ConfigReader
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.enums import ConfigKey, Methods, Options, TagNames
from simba.utils.errors import SamplingError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import read_config_entry


class TrainMultiLabelRandomForestClassifier(ConfigReader, TrainModelMixin):
    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path)
        TrainModelMixin.__init__(self)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        self.read_model_settings_from_config(config=self.config)

        check_if_filepath_list_is_empty(
            filepaths=self.target_file_paths,
            error_msg="Zero annotation files found in project_folder/csv/targets_inserted, cannot create model.",
        )
        print(f"Reading in {len(self.target_file_paths)} annotated files...")
        self.data_df = self.read_all_files_in_folder_mp_futures(
            file_paths=self.target_file_paths,
            file_type=self.file_type,
            classifier_names=self.clf_names,
            raise_bool_clf_error=False,
        )
        self.check_raw_dataset_integrity(df=self.data_df, logs_path=self.logs_path)
        self.data_df_wo_cords = self.drop_bp_cords(df=self.data_df)
        annotation_cols_to_remove = self.read_in_all_model_names_to_remove(
            self.config, self.clf_cnt, self.clf_name
        )
        self.x_y_df = self.delete_other_annotation_columns(
            df=self.data_df_wo_cords,
            annotations_lst=list(annotation_cols_to_remove),
            raise_error=False,
        )
        self.classifier_map = ast.literal_eval(
            read_config_entry(
                self.config,
                ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,
                "classifier_map",
                data_type="str",
            )
        )
        self.x_df, self.y_df = self.split_df_to_x_y(self.x_y_df, self.clf_name)
        self.feature_names = self.x_df.columns
        self.check_sampled_dataset_integrity(x_df=self.x_df, y_df=self.y_df)
        print(f"Number of features in dataset: {len(self.feature_names)}")
        for k, v in self.classifier_map.items():
            print(
                f"Number of {v} frames in dataset: {len(self.y_df[self.y_df == k])} ({(len(self.y_df[self.y_df == k]) / len(self.x_df)) * 100}%)"
            )

    def _check_presence_of_classes_post_sampling(self):
        for set, clf_code in product(
            [self.y_train, self.y_test], self.classifier_map.keys()
        ):
            obs_cnt = len(self.y_df[self.y_df == clf_code])
            if obs_cnt == 0:
                raise SamplingError(
                    msg=f"Zero observations of {self.classifier_map[clf_code]} found in the training and/or test set."
                )

    def perform_sampling(self):
        if self.split_type == Methods.SPLIT_TYPE_FRAMES.value:
            self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(
                self.x_df, self.y_df, test_size=self.tt_size
            )
        self._check_presence_of_classes_post_sampling()

    def run(self):
        print("Training and evaluating model...")
        self.timer = SimbaTimer(start=True)
        self.rf_clf = RandomForestClassifier(
            n_estimators=self.n_estimators,
            max_features=self.max_features,
            n_jobs=-1,
            criterion=self.criterion,
            min_samples_leaf=self.min_sample_leaf,
            bootstrap=True,
            verbose=1,
            class_weight=self.class_weights,
        )
        self.rf_clf = self.clf_fit(clf=self.rf_clf, x_df=self.x_df, y_df=self.y_df)
        if self.compute_permutation_importance in Options.PERFORM_FLAGS.value:
            self.calc_permutation_importance(
                self.x_test,
                self.y_test,
                self.rf_clf,
                self.feature_names,
                self.clf_name,
                self.eval_out_path,
            )
        if self.generate_learning_curve in Options.PERFORM_FLAGS.value:
            self.calc_learning_curve(
                x_y_df=self.x_y_df,
                clf_name=self.clf_name,
                shuffle_splits=self.shuffle_splits,
                dataset_splits=self.dataset_splits,
                tt_size=self.tt_size,
                rf_clf=self.rf_clf,
                save_dir=self.eval_out_path,
            )
        if self.generate_precision_recall_curve in Options.PERFORM_FLAGS.value:
            self.calc_pr_curve(
                self.rf_clf, self.x_test, self.y_test, self.clf_name, self.eval_out_path
            )
        if self.generate_example_decision_tree in Options.PERFORM_FLAGS.value:
            self.create_example_dt(
                self.rf_clf,
                self.clf_name,
                self.feature_names,
                list(self.classifier_map.values()),
                self.eval_out_path,
            )
        if self.generate_classification_report in Options.PERFORM_FLAGS.value:
            self.create_clf_report(
                self.rf_clf,
                self.x_test,
                self.y_test,
                list(self.classifier_map.values()),
                self.eval_out_path,
            )
        if self.generate_features_importance_log in Options.PERFORM_FLAGS.value:
            self.create_x_importance_log(
                self.rf_clf, self.feature_names, self.clf_name, self.eval_out_path
            )
        if self.generate_features_importance_bar_graph in Options.PERFORM_FLAGS.value:
            self.create_x_importance_bar_chart(
                self.rf_clf,
                self.feature_names,
                self.clf_name,
                self.eval_out_path,
                self.feature_importance_bars,
            )
        if self.generate_example_decision_tree_fancy in Options.PERFORM_FLAGS.value:
            self.dviz_classification_visualization(
                self.x_train,
                self.y_train,
                self.clf_name,
                list(self.classifier_map.values()),
                self.eval_out_path,
            )


model_trainer = TrainMultiLabelRandomForestClassifier(
    config_path="/Users/simon/Desktop/envs/troubleshooting/multilabel/project_folder/project_config.ini"
)
model_trainer.perform_sampling()
model_trainer.run()



__author__ = "Simon Nilsson"

import glob
import os
from copy import deepcopy

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_if_filepath_list_is_empty,
                                check_that_column_exist)
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df


class SolomonImporter(ConfigReader):
    """
    Append SOLOMON human annotations onto featurized pose-estimation data.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str data_dir: path to folder holding SOLOMON data files is CSV format

    .. note::
       `Third-party import tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/third_party_annot.md>`__.
       `Example of expected SOLOMON file format <https://github.com/sgoldenlab/simba/blob/master/misc/solomon_example.csv>`__.

    Examples
    ----------
    >>> solomon_imported = SolomonImporter(config_path=r'MySimBAConfigPath', data_dir=r'MySolomonDir')
    >>> solomon_imported.run()

    References
    ----------

    .. [1] `SOLOMON CODER USER-GUIDE (PDF) <https://solomon.andraspeter.com/Solomon%20Intro.pdf>`__.
    """

    def __init__(self, config_path: str, data_dir: str):
        super().__init__(config_path=config_path)
        self.solomon_paths = glob.glob(data_dir + "/*.csv")
        check_if_filepath_list_is_empty(
            filepaths=self.solomon_paths,
            error_msg=f"SIMBA ERROR: No CSV files detected in SOLOMON directory {data_dir}",
        )
        check_if_filepath_list_is_empty(
            filepaths=self.feature_file_paths,
            error_msg=f"SIMBA ERROR: No CSV files detected in feature directory {self.features_dir}",
        )
        if not os.path.exists(self.targets_folder):
            os.mkdir(self.targets_folder)

    def run(self):
        for file_cnt, file_path in enumerate(self.solomon_paths):
            _, file_name, _ = get_fn_ext(file_path)
            feature_file_path = os.path.join(
                self.features_dir, file_name + "." + self.file_type
            )
            _, _, fps = self.read_video_info(video_name=file_name)
            if not os.path.isfile(feature_file_path):
                print(
                    "SIMBA WARNING: Data for video {} does not exist in the features directory. SimBA will SKIP appending annotations for video {}".format(
                        file_name, file_name
                    )
                )
                continue
            save_path = os.path.join(
                self.targets_folder, file_name + "." + self.file_type
            )
            solomon_df = pd.read_csv(file_path)
            check_that_column_exist(
                df=solomon_df, column_name="Behaviour", file_name=file_path
            )
            features_df = read_df(feature_file_path, self.file_type)
            out_df = deepcopy(features_df)
            features_frames = list(features_df.index)
            solomon_df["frame_cnt"] = solomon_df.index
            for clf_name in self.clf_names:
                target_col = list(solomon_df.columns[solomon_df.isin([clf_name]).any()])
                if len(target_col) == 0:
                    print(
                        "SIMBA WARNING: No SOLOMON frames annotated as containing behavior {} in video {}. SimBA will set all frames in video {} as behavior-absent for behavior {}".format(
                            clf_name, file_name, file_name, clf_name
                        )
                    )
                    continue
                target_frm_list = list(
                    solomon_df["frame_cnt"][solomon_df[target_col[0]] == clf_name]
                )
                idx_difference = list(set(target_frm_list) - set(features_frames))
                if len(idx_difference) > 0:
                    if len(idx_difference) > 0:
                        print(
                            f"SIMBA SOLOMON WARNING: SimBA found SOLOMON annotations for behavior {clf_name} in video "
                            f"{file_name} that are annotated to occur at times which is not present in the "
                            f"video data you imported into SIMBA. The video you imported to SimBA has {str(features_frames[-1])} frames. "
                            f"However, in SOLOMON, you have annotated {clf_name} to happen at frame number {str(idx_difference[0])}. "
                            f"These ambiguous annotations occur in {str(len(idx_difference))} different frames for video {file_name} that SimBA will **remove** by default. "
                            f"Please make sure you imported the same video as you annotated in SOLOMON into SimBA and the video is registered with the correct frame rate."
                        )
                    target_frm_list = [
                        x for x in target_frm_list if x not in idx_difference
                    ]
                out_df[clf_name] = 0
                out_df.loc[target_frm_list, clf_name] = 1

            write_df(out_df, self.file_type, save_path)
            print("Solomon annotations appended for video {}...".format(file_name))
        stdout_success(
            msg="All SOLOMON annotations imported. Data saved in the project_folder/csv/targets_inserted directory of the SimBA project"
        )


# test = SolomonImporter(
#     config_path="/Users/simon/Desktop/envs/simba_dev/test/data/test_projects/two_c57/project_folder/project_config.ini",
#     data_dir="/Users/simon/Desktop/envs/simba_dev/test/data/test_projects/two_c57/solomon_annotations",
# )
#
# test.run()



import os
from copy import deepcopy
from typing import Dict

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.errors import InvalidInputError, MissingColumnsError
from simba.utils.printing import stdout_success
from simba.utils.read_write import (get_fn_ext, read_df, read_video_info,
                                    str_2_bool)


class BooleanConditionalCalculator(ConfigReader):
    """
    Compute descriptive statistics (e.g., the time in seconds and number of frames) of multiple Boolean fields fullfilling user-defined conditions.

    For example, computedescriptive statistics for when Animal 1 is inside the shape Rectangle_1 while at the same time directing towards shape Polygon_1,
    while at the same time Animal 2 is outside shape Rectangle_1 and directing towards Polygon_1.

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter Dict[str, bool] rules: Rules with field names as keys and bools as values.

    .. note:
       `Example expected output table <https://github.com/sgoldenlab/simba/blob/master/misc/Conditional_aggregate_statistics_20231004130314.csv>`__.

    Examples
    -----
    >>> rules = {'Rectangle_1 Simon in zone': 'TRUE', 'Polygon_1 JJ in zone': 'TRUE'}
    >>> conditional_bool_rule_calculator = BooleanConditionalCalculator(rules=rules, config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini')
    >>> conditional_bool_rule_calculator.run()
    >>> conditional_bool_rule_calculator.save()
    """

    def __init__(self, config_path: str, rules: Dict[str, bool]):
        ConfigReader.__init__(self, config_path=config_path)
        self.save_path = os.path.join(
            self.logs_path, f"Conditional_aggregate_statistics_{self.datetime}.csv"
        )
        check_if_filepath_list_is_empty(
            filepaths=self.feature_file_paths,
            error_msg=f"No data found in {self.features_dir}",
        )
        self.rules = rules
        self.output_df = pd.DataFrame(
            columns=["VIDEO"] + list(self.rules.keys()) + ["TIME (s)", "FRAMES (count)"]
        )

    def _check_integrity_of_rule_columns(self):
        for behavior in self.rules.keys():
            if behavior not in self.df.columns:
                raise MissingColumnsError(
                    msg=f"File is missing the column {behavior} which is required for your conditional aggregate statistics {self.file_path}",
                    source=self.__class__.__name__,
                )
            other_col_vals = [
                x for x in list(set(self.df[behavior])) if x not in [0, 1]
            ]
            if len(other_col_vals) > 0:
                raise InvalidInputError(
                    msg=f"Invalid values found in column {behavior} of file {self.file_path}: SimBA expects only 0 and 1s.",
                    source=self.__class__.__name__,
                )

    def run(self):
        for file_cnt, file_path in enumerate(self.feature_file_paths):
            self.file_path = file_path
            _, self.video_name, _ = get_fn_ext(filepath=file_path)
            _, _, self.fps = read_video_info(
                vid_info_df=self.video_info_df, video_name=self.video_name
            )
            self.df = read_df(file_path=file_path, file_type=self.file_type)
            self._check_integrity_of_rule_columns()
            self.df = self.df[list(self.rules.keys())]

            self.sliced_df = deepcopy(self.df)
            values_str = []
            for k, v in self.rules.items():
                if str_2_bool(v):
                    self.sliced_df = self.sliced_df[self.sliced_df[k] == 1]
                else:
                    self.sliced_df = self.sliced_df[self.sliced_df[k] == 0]
                values_str.append(v)
            time_s = round(len(self.sliced_df) / self.fps, 4)
            self.output_df.loc[len(self.output_df)] = (
                [self.video_name]
                + list(self.rules.values())
                + [time_s]
                + [len(self.sliced_df)]
            )

    def save(self):
        self.output_df.to_csv(self.save_path, index=False)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Boolean conditional data saved at at {self.save_path}!",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


# rules = {'Rectangle_1 Simon in zone': 'TRUE', 'Polygon_1 JJ in zone': 'TRUE'}
# runner = BooleanConditionalCalculator(rules=rules, config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini')
# runner.run()
# runner.save()



import numpy as np
from numba import njit, prange
from numba.np.extensions import cross2d


@njit("(float64[:,:], int64[:], int64, int64)", cache=True)
def process(S, P, a, b):
    signed_dist = cross2d(S[P] - S[a], S[b] - S[a])
    K = np.array(
        [i for s, i in zip(signed_dist, P) if s > 0 and i != a and i != b],
        dtype=np.int64,
    )
    if len(K) == 0:
        return [a, b]
    c = P[np.argmax(signed_dist)]
    return process(S, K, a, c)[:-1] + process(S, K, c, b)


@njit("(float64[:, :,:],)", cache=True, fastmath=True)
def quickhull_2d(points: np.ndarray) -> np.ndarray:
    results = np.full((points.shape[0]), np.nan)

    def Perimeter(xy):
        perimeter = np.linalg.norm(xy[0] - xy[-1])
        for i in prange(xy.shape[0] - 1):
            p = np.linalg.norm(xy[i] - xy[i + 1])
            perimeter += p
        return perimeter

    for i in range(points.shape[0]):
        S = points[i, :, :]
        a, b = np.argmin(S[:, 0]), np.argmax(S[:, 0])
        max_index = np.argmax(S[:, 0])
        idx = (
            process(S, np.arange(S.shape[0]), a, max_index)[:-1]
            + process(S, np.arange(S.shape[0]), max_index, a)[:-1]
        )
        x, y = np.full((len(idx)), np.nan), np.full((len(idx)), np.nan)
        for j in prange(len(idx)):
            x[j], y[j] = S[idx[j], 0], S[idx[j], 1]

        x0, y0 = np.mean(x), np.mean(y)
        r = np.sqrt((x - x0) ** 2 + (y - y0) ** 2)
        angles = np.where(
            (y - y0) > 0, np.arccos((x - x0) / r), 2 * np.pi - np.arccos((x - x0) / r)
        )
        mask = np.argsort(angles)
        x_sorted, y_sorted = x[mask], y[mask]
        xy = np.vstack((x_sorted, y_sorted)).T
        results[i] = Perimeter(xy)

    return results


points = np.random.randint(1, 50, size=(50000, 5, 2)).astype(float)
start = time.time()
results = quickhull_2d(points)
print(time.time() - start)
#
points = points.astype(int)
start = time.time()
results = np.full((points.shape[0]), np.nan)
for i in range(points.shape[0]):
    results[i] = ConvexHull(points[i], qhull_options="En").area
print(time.time() - start)
# #


# print(ch) # print [0, 4, 1, 3]



__author__ = "Simon Nilsson"

import glob
import os

import numpy as np
import pandas as pd
from numba import jit, prange

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.enums import ConfigKey, Dtypes
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (get_fn_ext, read_config_entry, read_df,
                                    write_df)


class OutlierCorrecterMovement(ConfigReader, FeatureExtractionMixin):
    """
    Detect and ammend outliers in pose-estimation data based on movement lenghth (Euclidean) of the body-parts
    in the current frame from preceeding frame. Uses critera stored in the SimBA project project_config.ini
    under the [Outlier settings] header.

    :parameter str config_path: path to SimBA project config file in Configparser format

    .. image:: _static/img/movement_outlier.png
       :width: 500
       :align: center

    .. note::
       `Outlier correction documentation <https://github.com/sgoldenlab/simba/blob/master/misc/Outlier_settings.pdf>`__.

    Examples
    ----------
    >>> outlier_correcter_movement = OutlierCorrecterMovement(config_path='MyProjectConfig')
    >>> outlier_correcter_movement.run()
    """

    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        FeatureExtractionMixin.__init__(self)
        if not os.path.exists(self.outlier_corrected_movement_dir):
            os.makedirs(self.outlier_corrected_movement_dir)
        if self.animal_cnt == 1:
            self.animal_id = read_config_entry(
                self.config,
                ConfigKey.MULTI_ANIMAL_ID_SETTING.value,
                ConfigKey.MULTI_ANIMAL_IDS.value,
                Dtypes.STR.value,
            )
            if self.animal_id != "None":
                self.animal_bp_dict[self.animal_id] = self.animal_bp_dict.pop(
                    "Animal_1"
                )
        self.above_criterion_dict_dict = {}
        self.criterion = read_config_entry(
            self.config,
            ConfigKey.OUTLIER_SETTINGS.value,
            ConfigKey.MOVEMENT_CRITERION.value,
            Dtypes.FLOAT.value,
        )
        self.outlier_bp_dict = {}
        for animal_name in self.animal_bp_dict.keys():
            self.outlier_bp_dict[animal_name] = {}
            self.outlier_bp_dict[animal_name]["bp_1"] = read_config_entry(
                self.config,
                "Outlier settings",
                "movement_bodypart1_{}".format(animal_name.lower()),
                "str",
            )
            self.outlier_bp_dict[animal_name]["bp_2"] = read_config_entry(
                self.config,
                "Outlier settings",
                "movement_bodypart2_{}".format(animal_name.lower()),
                "str",
            )

    @staticmethod
    @jit(nopython=True)
    def __corrector(data=np.ndarray, criterion=float):
        results, current_value, cnt = np.full(data.shape, np.nan), data[0, :], 0
        for i in range(data.shape[0]):
            dist = abs(np.linalg.norm(current_value - data[i, :]))
            if dist <= criterion:
                current_value = data[i, :]
                cnt += 1
            results[i, :] = current_value
        return results, cnt

    def __outlier_replacer(self):
        for animal_name, animal_body_parts in self.animal_bp_dict.items():
            for bp_x_name, bp_y_name in zip(
                animal_body_parts["X_bps"], animal_body_parts["Y_bps"]
            ):
                vals, cnt = self.__corrector(
                    data=self.data_df[[bp_x_name, bp_y_name]].values,
                    criterion=self.animal_criteria[animal_name],
                )
                df = pd.DataFrame(vals, columns=[bp_x_name, bp_y_name])
                self.data_df.update(df)
                self.log.loc[len(self.log)] = [
                    self.video_name,
                    animal_name,
                    bp_x_name[:-2],
                    cnt,
                    round(cnt / len(df), 6),
                ]

    def run(self):
        """
        Runs outlier detection and correction. Results are stored in the
        ``project_folder/csv/outlier_corrected_movement`` directory of the SimBA project.
        """
        self.log = pd.DataFrame(
            columns=[
                "VIDEO",
                "ANIMAL",
                "BODY-PART",
                "CORRECTION COUNT",
                "CORRECTION PCT",
            ]
        )
        for file_cnt, file_path in enumerate(self.input_csv_paths):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            print(
                "Processing video {}. Video {}/{}...".format(
                    self.video_name, str(file_cnt + 1), str(len(self.input_csv_paths))
                )
            )
            self.above_criterion_dict_dict[self.video_name] = {}
            save_path = os.path.join(
                self.outlier_corrected_movement_dir,
                self.video_name + "." + self.file_type,
            )
            self.data_df = read_df(file_path, self.file_type, check_multiindex=True)
            self.data_df = self.insert_column_headers_for_outlier_correction(
                data_df=self.data_df, new_headers=self.bp_headers, filepath=file_path
            )
            self.data_df_combined = self.create_shifted_df(df=self.data_df)
            self.animal_criteria = {}
            for animal_name, animal_bps in self.outlier_bp_dict.items():
                animal_bp_distances = np.sqrt(
                    (
                        self.data_df[animal_bps["bp_1"] + "_x"]
                        - self.data_df[animal_bps["bp_2"] + "_x"]
                    )
                    ** 2
                    + (
                        self.data_df[animal_bps["bp_1"] + "_y"]
                        - self.data_df[animal_bps["bp_2"] + "_y"]
                    )
                    ** 2
                )
                self.animal_criteria[animal_name] = (
                    animal_bp_distances.mean() * self.criterion
                )
            self.__outlier_replacer()
            write_df(df=self.data_df, file_type=self.file_type, save_path=save_path)
            video_timer.stop_timer()
            print(
                f"Corrected movement outliers for file {self.video_name} (elapsed time: {video_timer.elapsed_time_str}s)..."
            )
        self.__save_log_file()

    def __save_log_file(self):
        self.log_fn = os.path.join(
            self.logs_path, "Outliers_movement_{}.csv".format(self.datetime)
        )
        self.log.to_csv(self.log_fn)
        self.timer.stop_timer()
        stdout_success(
            msg='Log for corrected "movement outliers" saved in project_folder/logs',
            elapsed_time=self.timer.elapsed_time_str,
        )


#
# test = OutlierCorrecterMovement(config_path='/Users/simon/Desktop/envs/troubleshooting/dorian_2/project_folder/project_config.ini')
# test.run()

# test = OutlierCorrecterMovement(config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini')
# test.correct_movement_outliers()

# test = OutlierCorrecterMovement(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini')
# test.run()

# test = OutlierCorrecterMovement(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.correct_movement_outliers()
#
# test = OutlierCorrecterMovement(config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini')
# test.correct_movement_outliers()



__author__ = "Simon Nilsson"

import os
from typing import List

import cv2
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from numba import jit, prange

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.enums import Formats
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class HeatMapperClfSingleCore(ConfigReader, PlottingMixin):
    """
    Create heatmaps representing the locations of the classified behavior.

    .. note::
       `GitHub visualizations tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.
       For improved run-time, see :meth:`simba.heat_mapper_clf_mp.HeatMapperClfMultiprocess` for multiprocess class.

    .. image:: _static/img/heatmap.png
       :width: 300
       :align: center

    :param str config_path: path to SimBA project config file in Configparser format
    :param bool final_img_setting: If True, then create a single image representing the last frame of the input video
    :param bool video_setting: If True, then create a video of heatmaps.
    :param bool frame_setting: If True, then create individual heatmap frames.
    :param int bin_size: The rectangular size of each heatmap location in millimeters. For example, `50` will divide the video into
        5 centimeter rectangular spatial bins.
    :param str palette: Heatmap pallette. Eg. 'jet', 'magma', 'inferno','plasma', 'viridis', 'gnuplot2'
    :param str bodypart: The name of the body-part used to infer the location of the classified behavior
    :param str clf_name: The name of the classified behavior.
    :param int or 'auto' max_scale: The max value in the heatmap in seconds. E.g., with a value of `10`, if the classified behavior has occured
        >= 10 within a rectangular bins, it will be filled with the same color.

    Examples
    -----
    >>> heat_mapper_clf = HeatMapperClfSingleCore(config_path='MyConfigPath', final_img_setting=False, video_setting=True, frame_setting=False, bin_size=50, palette='jet', bodypart='Nose_1', clf_name='Attack', max_scale=20).run()

    """

    def __init__(
        self,
        config_path: str,
        final_img_setting: bool,
        video_setting: bool,
        frame_setting: bool,
        bodypart: str,
        clf_name: str,
        files_found: List[str],
        style_attr: dict,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)

        if (not frame_setting) and (not video_setting) and (not final_img_setting):
            raise NoSpecifiedOutputError(
                msg="Please choose to select either heatmap videos, frames, and/or final image."
            )
        self.frame_setting, self.video_setting = frame_setting, video_setting
        self.final_img_setting, self.bp = final_img_setting, bodypart
        self.bin_size, self.max_scale, self.palette, self.shading = (
            style_attr["bin_size"],
            style_attr["max_scale"],
            style_attr["palette"],
            style_attr["shading"],
        )
        self.clf_name, self.files_found = clf_name, files_found
        if not os.path.exists(self.heatmap_clf_location_dir):
            os.makedirs(self.heatmap_clf_location_dir)
        self.bp_lst = [self.bp + "_x", self.bp + "_y"]
        self.timer = SimbaTimer(start=True)
        print("Processing {} video(s)...".format(str(len(self.files_found))))

    @staticmethod
    @jit(nopython=True)
    def __calculate_cum_array(clf_array: np.array, fps: int):
        cum_sum_arr = np.full(clf_array.shape, np.nan)
        for frm_idx in prange(clf_array.shape[0]):
            frame_cum_sum = np.full((clf_array.shape[1], clf_array.shape[2]), 0.0)
            sliced_arr = clf_array[0:frm_idx]
            for i in range(sliced_arr.shape[0]):
                for j in range(sliced_arr.shape[1]):
                    for k in range(sliced_arr.shape[2]):
                        frame_cum_sum[j][k] += sliced_arr[i][j][k]
            cum_sum_arr[frm_idx] = frame_cum_sum

        return cum_sum_arr / fps

    def __calculate_bin_attr(
        self,
        data_df: pd.DataFrame,
        clf_name: str,
        bp_lst: list,
        px_per_mm: int,
        img_width: int,
        img_height: int,
        bin_size: int,
        fps: int,
    ):
        bin_size_px = int(float(px_per_mm) * float(bin_size))
        horizontal_bin_cnt = int(img_width / bin_size_px)
        vertical_bin_cnt = int(img_height / bin_size_px)
        aspect_ratio = round((vertical_bin_cnt / horizontal_bin_cnt), 3)

        clf_idx = (
            data_df[bp_lst][data_df[clf_name] == 1].reset_index().to_numpy().astype(int)
        )

        bin_dict = {}
        x_location, y_location = 0, 0
        for hbin in range(horizontal_bin_cnt):
            bin_dict[hbin] = {}
            for vbin in range(vertical_bin_cnt):
                bin_dict[hbin][vbin] = {
                    "top_left_x": x_location,
                    "top_left_y": y_location,
                    "bottom_right_x": x_location + bin_size_px,
                    "bottom_right_y": y_location + bin_size_px,
                }
                y_location += bin_size_px
            y_location = 0
            x_location += bin_size_px

        clf_array = np.zeros((len(data_df), vertical_bin_cnt, horizontal_bin_cnt))

        for clf_frame in clf_idx:
            for h_bin_name, v_dict in bin_dict.items():
                for v_bin_name, c in v_dict.items():
                    if (
                        clf_frame[1] <= c["bottom_right_x"]
                        and clf_frame[1] >= c["top_left_x"]
                    ):
                        if (
                            clf_frame[2] <= c["bottom_right_y"]
                            and clf_frame[2] >= c["top_left_y"]
                        ):
                            clf_array[int(clf_frame[0])][v_bin_name][h_bin_name] = 1

        clf_array = self.__calculate_cum_array(clf_array=clf_array, fps=fps)

        return clf_array, aspect_ratio

    def __calculate_max_scale(self, clf_array: np.array):
        return np.round(np.max(np.max(clf_array[-1], axis=0)), 3)

    def run(self):
        """
        Creates heatmap charts. Results are stored in the `project_folder/frames/heatmaps_classifier_locations`
        directory of SimBA project.

        Returns
        ----------
        None
        """

        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer()
            video_timer.start_timer()
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_info, self.px_per_mm, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            self.width, self.height = int(
                self.video_info["Resolution_width"].values[0]
            ), int(self.video_info["Resolution_height"].values[0])
            if self.video_setting:
                self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
                self.video_save_path = os.path.join(
                    self.heatmap_clf_location_dir, self.video_name + ".mp4"
                )
                self.writer = cv2.VideoWriter(
                    self.video_save_path,
                    self.fourcc,
                    self.fps,
                    (self.width, self.height),
                )
            if self.frame_setting:
                self.save_video_folder = os.path.join(
                    self.heatmap_clf_location_dir, self.video_name
                )
                if not os.path.exists(self.save_video_folder):
                    os.makedirs(self.save_video_folder)
            self.data_df = read_df(file_path=file_path, file_type=self.file_type)
            clf_array, aspect_ratio = self.__calculate_bin_attr(
                data_df=self.data_df,
                clf_name=self.clf_name,
                bp_lst=self.bp_lst,
                px_per_mm=self.px_per_mm,
                img_width=self.width,
                img_height=self.height,
                bin_size=self.bin_size,
                fps=self.fps,
            )

            if self.max_scale == "auto":
                self.max_scale = self.__calculate_max_scale(clf_array=clf_array)
                if self.max_scale == 0:
                    self.max_scale = 1

            if self.final_img_setting:
                self.make_clf_heatmap_plot(
                    frm_data=clf_array[-1, :, :],
                    max_scale=self.max_scale,
                    palette=self.palette,
                    aspect_ratio=aspect_ratio,
                    file_name=os.path.join(
                        self.heatmap_clf_location_dir,
                        self.video_name + "_final_frm.png",
                    ),
                    shading=self.shading,
                    clf_name=self.clf_name,
                    img_size=(self.width, self.height),
                    final_img=True,
                )

            if self.video_setting or self.frame_setting:
                for frm_cnt, cumulative_frm_idx in enumerate(range(clf_array.shape[0])):
                    frm_data = clf_array[cumulative_frm_idx, :, :]
                    cum_df = pd.DataFrame(frm_data).reset_index()
                    cum_df = cum_df.melt(
                        id_vars="index",
                        value_vars=None,
                        var_name=None,
                        value_name="seconds",
                        col_level=None,
                    ).rename(
                        columns={"index": "vertical_idx", "variable": "horizontal_idx"}
                    )
                    cum_df["color"] = (
                        (cum_df["seconds"].astype(float) / float(self.max_scale))
                        .round(2)
                        .clip(upper=100)
                    )
                    color_array = np.zeros(
                        (
                            len(cum_df["vertical_idx"].unique()),
                            len(cum_df["horizontal_idx"].unique()),
                        )
                    )
                    for i in range(color_array.shape[0]):
                        for j in range(color_array.shape[1]):
                            value = cum_df["color"][
                                (cum_df["horizontal_idx"] == j)
                                & (cum_df["vertical_idx"] == i)
                            ].values[0]
                            color_array[i, j] = value

                    fig = plt.figure()
                    im_ratio = color_array.shape[0] / color_array.shape[1]
                    plt.pcolormesh(
                        color_array,
                        shading=self.shading,
                        cmap=self.palette,
                        rasterized=True,
                        alpha=1,
                        vmin=0.0,
                        vmax=float(self.max_scale),
                    )
                    plt.gca().invert_yaxis()
                    plt.xticks([])
                    plt.yticks([])
                    plt.axis("off")
                    plt.tick_params(axis="both", which="both", length=0)
                    cb = plt.colorbar(pad=0.0, fraction=0.023 * im_ratio)
                    cb.ax.tick_params(size=0)
                    cb.outline.set_visible(False)
                    cb.set_label(
                        "{} (seconds)".format(self.clf_name), rotation=270, labelpad=10
                    )
                    plt.tight_layout()
                    plt.gca().set_aspect(aspect_ratio)
                    canvas = FigureCanvas(fig)
                    canvas.draw()
                    mat = np.array(canvas.renderer._renderer)
                    image = cv2.cvtColor(mat, cv2.COLOR_RGB2BGR)
                    image = cv2.resize(image, (self.width, self.height))
                    image = np.uint8(image)
                    plt.close()

                    if self.video_setting:
                        self.writer.write(image)
                    if self.frame_setting:
                        frame_save_path = os.path.join(
                            self.save_video_folder, str(frm_cnt) + ".png"
                        )
                        cv2.imwrite(frame_save_path, image)
                    print(
                        "Created heatmap frame: {} / {}. Video: {} ({}/{})".format(
                            str(frm_cnt + 1),
                            str(len(self.data_df)),
                            self.video_name,
                            str(file_cnt + 1),
                            len(self.files_found),
                        )
                    )

                if self.video_setting:
                    self.writer.release()

                video_timer.stop_timer()
                print(
                    "Heatmap plot for video {} saved (elapsed time: {}s) ... ".format(
                        self.video_name, video_timer.elapsed_time_str
                    )
                )

        self.timer.stop_timer()
        stdout_success(
            msg="All heatmap visualizations created in project_folder/frames/output/heatmaps_classifier_locations directory",
            elapsed_time="self.timer.elapsed_time_str",
        )


# test = HeatMapperClfSingleCore(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini',
#                      style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 75, 'max_scale': 'auto'},
#                      final_img_setting=False,
#                      video_setting=True,
#                      frame_setting=False,
#                      bodypart='Nose_1',
#                      clf_name='Attack',
#                      files_found=['/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/csv/machine_results/Together_3.csv'])
# test.run()



__author__ = "Simon Nilsson"

import os
from tkinter import *

from simba.data_processors.pup_retrieval_calculator import \
    PupRetrieverCalculator
from simba.roi_tools.ROI_analyzer import ROIAnalyzer
from simba.ui.tkinter_functions import DropDownMenu, Entry_Box, hxtScrollbar
from simba.utils.checks import check_float, check_int
from simba.utils.enums import ConfigKey, Formats, Paths
from simba.utils.errors import NoFilesFoundError
from simba.utils.read_write import (get_all_clf_names, read_config_entry,
                                    read_config_file)


class PupRetrievalPopUp(object):
    def __init__(self, config_path: str):
        self.smoothing_options, self.config_path = ["gaussian"], config_path
        self.smooth_factor_options = list(range(1, 11))
        self.config = read_config_file(config_path=config_path)
        self.project_path = read_config_entry(
            self.config,
            ConfigKey.GENERAL_SETTINGS.value,
            ConfigKey.PROJECT_PATH.value,
            data_type=ConfigKey.FOLDER_PATH.value,
        )
        self.ROI_path = os.path.join(self.project_path, Paths.ROI_DEFINITIONS.value)
        if not os.path.isfile(self.ROI_path):
            raise NoFilesFoundError(
                msg=f"Requires ROI definitions: no file found at {self.ROI_path}",
                source=self.__class__.__name__,
            )

        self.roi_analyzer = ROIAnalyzer(ini_path=config_path, data_path=None)
        self.roi_analyzer.run()
        self.shape_names = self.roi_analyzer.shape_names
        self.animal_names = self.roi_analyzer.multi_animal_id_list
        self.clf_names = get_all_clf_names(
            config=self.config, target_cnt=self.roi_analyzer.clf_cnt
        )

        self.distance_plots_var = BooleanVar(value=True)
        self.swarm_plot_var = BooleanVar(value=True)
        self.log_var = BooleanVar(value=True)

        self.main_frm = Toplevel()
        self.main_frm.minsize(400, 400)
        self.main_frm.wm_title("SIMBA PUP RETRIEVAL PROTOCOL 1")
        self.main_frm = hxtScrollbar(self.main_frm)
        self.main_frm.pack(expand=True, fill=BOTH)

        self.pup_track_p_entry = Entry_Box(
            self.main_frm, "Tracking probability (PUP): ", "20"
        )
        self.dam_track_p_entry = Entry_Box(
            self.main_frm, "Tracking probability (DAM): ", "20"
        )
        self.start_distance_criterion_entry = Entry_Box(
            self.main_frm, "Start distance criterion (MM):", "20", validation="numeric"
        )
        self.carry_frames_entry = Entry_Box(
            self.main_frm, "Carry time (S)", "20", validation="numeric"
        )
        self.core_nest_name_dropdown = DropDownMenu(
            self.main_frm, "Core-nest name: ", self.shape_names, "20"
        )
        self.nest_name_dropdown = DropDownMenu(
            self.main_frm, "Nest name: ", self.shape_names, "20"
        )
        self.dam_name_dropdown = DropDownMenu(
            self.main_frm, "Dam name: ", self.animal_names, "20"
        )
        self.pup_name_dropdown = DropDownMenu(
            self.main_frm, "Pup name: ", self.animal_names, "20"
        )
        self.smooth_function_dropdown = DropDownMenu(
            self.main_frm, "Smooth function: ", self.smoothing_options, "20"
        )
        self.smooth_factor_dropdown = DropDownMenu(
            self.main_frm, "Smooth factor: ", self.smooth_factor_options, "20"
        )
        self.max_time_entry = Entry_Box(
            self.main_frm, "Max time (S)", "20", validation="numeric"
        )
        self.carry_classifier_dropdown = DropDownMenu(
            self.main_frm, "Carry classifier name: ", self.clf_names, "20"
        )
        self.approach_classifier_dropdown = DropDownMenu(
            self.main_frm, "Approach classifier name: ", self.clf_names, "20"
        )
        self.dig_classifier_dropdown = DropDownMenu(
            self.main_frm, "Dig classifier name: ", self.clf_names, "20"
        )
        self.create_distance_plots_cb = Checkbutton(
            self.main_frm,
            text="Create distance plots (pre- and post tracking smoothing",
            variable=self.distance_plots_var,
        )
        self.swarm_plot_cb = Checkbutton(
            self.main_frm,
            text="Create results swarm plot",
            variable=self.swarm_plot_var,
        )
        self.log_cb = Checkbutton(
            self.main_frm, text="Create log-file", variable=self.log_var
        )

        self.pup_track_p_entry.entry_set(0.025)
        self.dam_track_p_entry.entry_set(0.5)
        self.start_distance_criterion_entry.entry_set(80)
        self.carry_frames_entry.entry_set(3)
        self.core_nest_name_dropdown.setChoices(choice=self.shape_names[0])
        self.nest_name_dropdown.setChoices(choice=self.shape_names[1])
        self.dam_name_dropdown.setChoices(choice=self.animal_names[0])
        self.pup_name_dropdown.setChoices(choice=self.animal_names[1])
        self.smooth_function_dropdown.setChoices(choice=self.smoothing_options[0])
        self.smooth_factor_dropdown.setChoices(choice=5)
        self.carry_frames_entry.entry_set(90)
        self.carry_classifier_dropdown.setChoices(self.clf_names[0])
        self.approach_classifier_dropdown.setChoices(self.clf_names[0])
        self.dig_classifier_dropdown.setChoices(self.clf_names[0])
        self.max_time_entry.entry_set(90)

        button_run = Button(
            self.main_frm,
            text="RUN",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="red",
            command=lambda: self.run(),
        )

        self.pup_track_p_entry.grid(row=0, sticky=W)
        self.dam_track_p_entry.grid(row=1, sticky=W)
        self.start_distance_criterion_entry.grid(row=2, sticky=W)
        self.carry_frames_entry.grid(row=3, sticky=W)
        self.core_nest_name_dropdown.grid(row=4, sticky=W)
        self.nest_name_dropdown.grid(row=5, sticky=W)
        self.dam_name_dropdown.grid(row=6, sticky=W)
        self.pup_name_dropdown.grid(row=7, sticky=W)
        self.smooth_function_dropdown.grid(row=8, sticky=W)
        self.smooth_factor_dropdown.grid(row=9, sticky=W)
        self.max_time_entry.grid(row=10, sticky=W)
        self.carry_classifier_dropdown.grid(row=11, sticky=W)
        self.approach_classifier_dropdown.grid(row=12, sticky=W)
        self.dig_classifier_dropdown.grid(row=13, sticky=W)
        self.swarm_plot_cb.grid(row=14, sticky=W)
        self.create_distance_plots_cb.grid(row=15, sticky=W)
        self.log_cb.grid(row=16, sticky=W)

        button_run.grid(row=17, sticky=W)

        # self.main_frm.mainloop()

    def run(self):
        pup_track_p = self.pup_track_p_entry.entry_get
        dam_track_p = self.dam_track_p_entry.entry_get
        start_distance_criterion = self.start_distance_criterion_entry.entry_get
        carry_frames = self.carry_frames_entry.entry_get
        core_nest = self.core_nest_name_dropdown.getChoices()
        nest = self.nest_name_dropdown.getChoices()
        dam_name = self.dam_name_dropdown.getChoices()
        pup_name = self.pup_name_dropdown.getChoices()
        smooth_function = self.smooth_function_dropdown.getChoices()
        smooth_factor = self.smooth_factor_dropdown.getChoices()
        max_time = self.max_time_entry.entry_get
        clf_carry = self.carry_classifier_dropdown.getChoices()
        clf_approach = self.approach_classifier_dropdown.getChoices()
        clf_dig = self.dig_classifier_dropdown.getChoices()
        check_float(
            name="Tracking probability (PUP)",
            value=pup_track_p,
            max_value=1.0,
            min_value=0.0,
        )
        check_float(
            name="Tracking probability (DAM)",
            value=dam_track_p,
            max_value=1.0,
            min_value=0.0,
        )
        check_float(
            name="Start distance criterion (MM)", value=start_distance_criterion
        )
        check_int(name="Carry frames (S)", value=carry_frames)
        check_int(name="max_time", value=max_time)

        swarm_plot = self.swarm_plot_var.get()
        distance_plot = self.distance_plots_var.get()
        log = self.log_var.get()

        settings = {
            "pup_track_p": float(pup_track_p),
            "dam_track_p": float(dam_track_p),
            "start_distance_criterion": float(start_distance_criterion),
            "carry_time": float(carry_frames),
            "core_nest": core_nest,
            "nest": nest,
            "dam_name": dam_name,
            "pup_name": pup_name,
            "smooth_function": smooth_function,
            "smooth_factor": int(smooth_factor),
            "max_time": float(max_time),
            "clf_carry": clf_carry,
            "clf_approach": clf_approach,
            "clf_dig": clf_dig,
            "swarm_plot": swarm_plot,
            "distance_plots": distance_plot,
            "log": log,
        }

        pup_calculator = PupRetrieverCalculator(
            config_path=self.config_path, settings=settings
        )
        pup_calculator.run()
        pup_calculator.save_results()



__author__ = "Simon Nilsson"

import os
from datetime import datetime
from typing import Dict, List, Union

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import (check_if_filepath_list_is_empty,
                                check_that_column_exist)
from simba.utils.data import create_color_palettes, detect_bouts, find_bins
from simba.utils.errors import InvalidVideoFileError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, get_video_meta_data, read_df
from simba.utils.warnings import NoDataFoundWarning


class SeverityBoutCalculator(ConfigReader, FeatureExtractionMixin):
    """
    Computes the "severity" of classification bout events based on how much
    the animals are moving within the bout. Bouts are scored as less or more severe at lower and higher movements, respectively.

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter dict settings: how to calculate the severity. E.g., {'brackets': 10, 'clf': 'Attack', 'animals': ['Simon', 'JJ'], 'normalization': 'ALL VIDEOS', 'save_bin_definitions': True, 'visualize': True, 'visualize_event_cnt': 'ALL', 'video_speed': 1.0, 'show_pose': True}

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md>`__.

    Examples
    ----------
    >>> settings = {'brackets': 10, 'clf': 'Attack', 'animals': ['Simon', 'JJ'], 'normalization': 'ALL VIDEOS', 'save_bin_definitions': True, 'visualize': True, 'visualize_event_cnt': 'ALL', 'video_speed': 1.0, 'show_pose': True}
    >>> processor = SeverityBoutCalculator(config_path='project_folder/project_config.ini', settings=settings)
    >>> processor.run()
    >>> processor.save()
    """

    def __init__(self, config_path: Union[str, os.PathLike], settings: Dict):
        ConfigReader.__init__(self, config_path=config_path)
        self.settings = settings
        self.movement_animal_bp_dict = {
            k: self.animal_bp_dict[k] for k in self.settings["animals"]
        }
        self.color_lst_lst = create_color_palettes(
            no_animals=len(list(self.animal_bp_dict.keys())),
            map_size=len(self.bp_headers),
        )
        check_if_filepath_list_is_empty(
            filepaths=self.machine_results_paths,
            error_msg=f"SIMBA ERROR: Cannot process severity. {self.machine_results_dir} directory is empty",
        )
        save_name = os.path.join(
            f'severity_{datetime.now().strftime("%Y%m%d%H%M%S")}.csv'
        )
        definitions_save_name = os.path.join(
            f'severity_bin_definitions_{datetime.now().strftime("%Y%m%d%H%M%S")}.csv'
        )
        self.save_path = os.path.join(self.logs_path, save_name)
        self.definitions_path, self.fourcc = os.path.join(
            self.logs_path, definitions_save_name
        ), cv2.VideoWriter_fourcc(*"mp4v")

    def __calculate_movements(self, data_paths: List[str]):
        movements = {}
        for file_cnt, file_path in enumerate(data_paths):
            _, video_name, _ = get_fn_ext(file_path)
            print(
                f"Analyzing movements in {video_name} ({file_cnt+1}/{len(data_paths)})..."
            )
            _, px_per_mm, fps = self.read_video_info(video_name=video_name)
            df = read_df(file_path=file_path, file_type=self.file_type)
            if self.settings["clf"] not in df.columns:
                NoDataFoundWarning(
                    msg=f'Skipping file {video_name} - {self.settings["clf"]} data not present in file'
                )
                continue
            video_movement = np.full((len(df)), 0)
            for animal_name, animal_bodyparts in self.movement_animal_bp_dict.items():
                animal_df = df[animal_bodyparts["X_bps"] + animal_bodyparts["Y_bps"]]
                animal_df = self.create_shifted_df(df=animal_df)
                for bp_x, bp_y in zip(
                    animal_bodyparts["X_bps"], animal_bodyparts["Y_bps"]
                ):
                    video_movement = np.add(
                        video_movement,
                        self.euclidean_distance(
                            animal_df[bp_x].values,
                            animal_df[f"{bp_x}_shifted"].values,
                            animal_df[bp_y].values,
                            animal_df[f"{bp_y}_shifted"].values,
                            px_per_mm,
                        ),
                    )
            movements[video_name] = video_movement
        return movements

    def run(self):
        movements = self.__calculate_movements(data_paths=self.machine_results_paths)
        video_bins_info = find_bins(
            data=movements,
            bracket_type=self.settings["bracket_type"],
            bracket_cnt=self.settings["brackets"],
            normalization_method=self.settings["normalization"],
        )
        self.results = pd.DataFrame(
            columns=[
                "VIDEO",
                "EVENT",
                "START TIME",
                "END TIME",
                "START FRAME",
                "END FRAME",
                "BOUT TIME",
                "MOVEMENT",
                "SEVERITY",
            ]
        )
        for file_cnt, file_path in enumerate(self.machine_results_paths):
            _, video_name, _ = get_fn_ext(file_path)
            _, px_per_mm, fps = self.read_video_info(video_name=video_name)
            df = read_df(file_path=file_path, file_type=self.file_type)
            check_that_column_exist(
                df=df, column_name=self.settings["clf"], file_name=file_path
            )
            bout_df = detect_bouts(
                data_df=df, target_lst=[self.settings["clf"]], fps=fps
            )
            bout_df.columns = [
                "EVENT",
                "START TIME",
                "END TIME",
                "START FRAME",
                "END FRAME",
                "BOUT TIME",
            ]
            severity_lst, movement_lst = [], []
            if len(bout_df) > 0:
                for i in bout_df[["START FRAME", "END FRAME"]].values:
                    bout_move_val = np.mean(movements[video_name][i[0] : i[1]])
                    if bout_move_val >= video_bins_info[video_name][-1][1]:
                        severity_lst.append(self.settings["brackets"])
                        movement_lst.append(bout_move_val)
                    else:
                        for j_cnt, j in enumerate(video_bins_info[video_name]):
                            if j[0] <= bout_move_val <= j[1]:
                                severity_lst.append(j_cnt + 1)
                                movement_lst.append(round(bout_move_val, 4))
                bout_df["MOVEMENT"] = movement_lst
                bout_df["SEVERITY"] = severity_lst
                bout_df.insert(0, "VIDEO", video_name)
                self.results = pd.concat([self.results, bout_df], axis=0).reset_index(
                    drop=True
                )
        self.save()
        if self.settings["save_bin_definitions"]:
            self.save_bin_definitions(data=video_bins_info)
        if self.settings["visualize"]:
            self.visualize()

    def save(self):
        self.results.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Bout severity data saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def save_bin_definitions(self, data=dict):
        results = pd.DataFrame(columns=["VIDEO", "SEVERITY_BIN", ">=", "<"])
        for video_name, video_bins in data.items():
            for bin_cnt, video_bin in enumerate(video_bins):
                results.loc[len(results)] = [
                    video_name,
                    bin_cnt + 1,
                    video_bin[0],
                    video_bin[1],
                ]
        save_path = os.path.join(
            self.logs_path, f"severity_bin_definitions_{self.datetime}.csv"
        )
        results.to_csv(save_path)
        stdout_success(msg=f"Severity bracket definitions saved at {save_path}")

    def visualize(self):
        if self.settings["visualize_event_cnt"] == "ALL CLIPS":
            self.settings["visualize_event_cnt"] = len(self.results)
        elif self.settings["visualize_event_cnt"] > len(self.results):
            self.settings["visualize_event_cnt"] = len(self.results)
            print(
                f'User specified {self.settings["visualize_event_cnt"]} visualization but only {len(self.results)} bouts where found. Creating {len(self.results)} videos...'
            )
        video_timer, save_dir = SimbaTimer(start=True), os.path.join(
            self.project_path, "frames", "output", "severity_bouts"
        )
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        viz_df = self.results.sample(
            n=self.settings["visualize_event_cnt"]
        ).reset_index(drop=True)
        for idx, r in viz_df.iterrows():
            video_name, start_frm, end_frame, severity = (
                r["VIDEO"],
                r["START FRAME"],
                r["END FRAME"],
                r["SEVERITY"],
            )
            pose_df = None
            if self.settings["show_pose"]:
                pose_df = read_df(
                    os.path.join(
                        self.machine_results_dir, video_name + "." + self.file_type
                    ),
                    self.file_type,
                    usecols=self.bp_headers,
                ).astype(int)
            clip_path = os.path.join(
                save_dir,
                f'{video_name}_{self.settings["clf"]}_{start_frm}_{end_frame}_{severity}.mp4',
            )
            video_path = self.find_video_of_file(
                video_dir=self.video_dir, filename=video_name
            )
            video_meta_data = get_video_meta_data(video_path)
            self.space_scale, self.radius_scale, self.res_scale, self.font_scale = (
                60,
                12,
                1500,
                1.1,
            )
            self.max_dim = max(video_meta_data["width"], video_meta_data["height"])
            self.circle_scale = int(self.radius_scale / (self.res_scale / self.max_dim))
            video_fps = int(video_meta_data["fps"] * self.settings["video_speed"])
            if video_fps < 1:
                video_fps = 1
            cap = cv2.VideoCapture(video_path)
            writer = cv2.VideoWriter(
                clip_path,
                cv2.VideoWriter_fourcc(*"mp4v"),
                video_fps,
                (int(video_meta_data["width"]), int(video_meta_data["height"])),
            )
            event_frm_count, frm_cnt, current_frm = end_frame - start_frm, 0, start_frm
            cap.set(1, current_frm)
            while current_frm < end_frame:
                ret, img = cap.read()
                if self.settings["show_pose"]:
                    frm_pose = pose_df.iloc[current_frm]
                    for animal_cnt, (animal_name, animal_body_parts) in enumerate(
                        self.animal_bp_dict.items()
                    ):
                        for bp_cnt, (x_name, y_name) in enumerate(
                            zip(animal_body_parts["X_bps"], animal_body_parts["Y_bps"])
                        ):
                            x = frm_pose[[x_name, y_name]].values
                            cv2.circle(
                                img,
                                (x[0], x[1]),
                                0,
                                self.color_lst_lst[animal_cnt][bp_cnt],
                                self.circle_scale,
                            )

                if not ret:
                    raise InvalidVideoFileError(
                        msg=f'Could not find frame {current_frm} in video {video_path}. Video {video_path} contains {video_meta_data["frame_count"]} frames.'
                    )
                writer.write(img)
                frm_cnt += 1
                print(
                    f"Frame {str(frm_cnt)} / {str(event_frm_count)}, Event {idx + 1}/{str(len(viz_df))}, Video {video_name}"
                )
                current_frm += 1
            writer.release()
        video_timer.stop_timer()
        stdout_success(
            msg=f"Videos complete: saved in {save_dir}",
            elapsed_time=video_timer.elapsed_time_str,
        )


# settings = {'brackets': 10,
#             'clf': 'Attack',
#             'animals': ['Simon', 'JJ'],
#             'normalization': 'ALL VIDEOS', #BY VIDEO
#             'bracket_type': "QUANTIZE",
#             'save_bin_definitions': True,
#             'visualize': True,
#             'visualize_event_cnt': 'ALL CLIPS',
#             'video_speed': 0.1,
#             'show_pose': True}
# processor = SeverityBoutCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', settings=settings)
# processor.run()



__author__ = "Simon Nilsson"


import functools
import itertools
import multiprocessing
import os
import platform

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.roi_tools.ROI_analyzer import ROIAnalyzer
from simba.utils.data import create_color_palettes, detect_bouts
from simba.utils.enums import Formats, Paths
from simba.utils.errors import DuplicationError, NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    get_video_meta_data)
from simba.utils.warnings import DuplicateNamesWarning

pd.options.mode.chained_assignment = None


class ROIPlotMultiprocess(ConfigReader, PlottingMixin):
    """
    Visualize the ROI data (number of entries/exits, time-spent-in etc).

    .. note::
       `ROI tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.
        .. image:: _static/img/roi_visualize.png
           :width: 400
           :align: center

    :param str config_path: Path to SimBA project config file in Configparser format
    :param str video_path: Name of video to create ROI visualizations for
    :param dict style_attr: User-defined visualization settings.
    :param int core_cnt: Number of cores to use.




    Examples
    ----------
    >>> roi_visualizer = ROIPlotMultiprocess(ini_path=r'MyProjectConfig', video_path="MyVideo.mp4")
    >>> roi_visualizer.run()
    """

    def __init__(self, ini_path: str, video_path: str, core_cnt: int, style_attr: dict):
        ConfigReader.__init__(self, config_path=ini_path)
        PlottingMixin.__init__(self)
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)
        self.roi_analyzer = ROIAnalyzer(
            ini_path=ini_path, data_path="outlier_corrected_movement_location"
        )
        self.video_path = os.path.join(self.video_dir, video_path)
        _, self.video_name, _ = get_fn_ext(video_path)
        self.core_cnt = core_cnt
        self.roi_analyzer.files_found = [
            os.path.join(
                self.roi_analyzer.input_folder,
                self.video_name + "." + self.roi_analyzer.file_type,
            )
        ]
        if not os.path.isfile(self.roi_analyzer.files_found[0]):
            raise NoFilesFoundError(
                msg=f"SIMBA ERROR: Could not find the file at path {self.roi_analyzer.files_found[0]}. Please make sure you have corrected body-part outliers or indicated that you want to skip outlier correction"
            )
        self.roi_analyzer.run()
        self.roi_entries_df = self.roi_analyzer.detailed_df
        self.data_df, self.style_attr = self.roi_analyzer.data_df, style_attr
        self.out_parent_dir = os.path.join(
            self.roi_analyzer.project_path, Paths.ROI_ANALYSIS.value
        )
        if not os.path.exists(self.out_parent_dir):
            os.makedirs(self.out_parent_dir)
        self.video_save_path = os.path.join(
            self.out_parent_dir, self.video_name + ".mp4"
        )
        self.video_shapes = list(
            itertools.chain(
                self.roi_analyzer.video_recs["Name"].unique(),
                self.roi_analyzer.video_circs["Name"].unique(),
                self.roi_analyzer.video_polys["Name"].unique(),
            )
        )
        if len(list(set(self.video_shapes))) != len(self.video_shapes):
            raise DuplicationError(
                msg="Some SIMBA ROI shapes have identical names. Please use unique names to visualize ROI data."
            )
        self.roi_analyzer.video_recs = self.add_missing_ROI_cols(
            self.roi_analyzer.video_recs
        )
        self.roi_analyzer.video_circs = self.add_missing_ROI_cols(
            self.roi_analyzer.video_circs
        )
        self.roi_analyzer.video_polys = self.add_missing_ROI_cols(
            self.roi_analyzer.video_polys
        )

        self.shape_dicts = {}
        for df in [
            self.roi_analyzer.video_recs,
            self.roi_analyzer.video_circs,
            self.roi_analyzer.video_polys,
        ]:
            if not df["Name"].is_unique:
                df = df.drop_duplicates(subset=["Name"], keep="first")
                DuplicateNamesWarning(
                    'Some of your ROIs with the same shape has the same names. E.g., you have two rectangles named "My rectangle". SimBA prefers ROI shapes with unique names. SimBA will keep one of the unique shape names and drop the rest.'
                )
            d = df.set_index("Name").to_dict(orient="index")
            self.shape_dicts = {**self.shape_dicts, **d}

        self.shape_columns = []
        for animal in self.roi_analyzer.multi_animal_id_list:
            for shape_name in self.video_shapes:
                self.data_df[animal + "_" + shape_name] = 0
                self.shape_columns.append(animal + "_" + shape_name)
        self.bp_dict = self.roi_analyzer.bp_dict
        self.output_folder = os.path.join(
            self.roi_analyzer.project_path, Paths.ROI_ANALYSIS.value
        )
        if not os.path.exists(self.output_folder):
            os.makedirs(self.output_folder)
        self.insert_data()

    def insert_data(self):
        self.roi_entries_dict = self.roi_entries_df[
            ["ANIMAL", "SHAPE", "ENTRY FRAMES", "EXIT FRAMES"]
        ].to_dict(orient="records")
        for entry_dict in self.roi_entries_dict:
            entry, exit = int(entry_dict["ENTRY FRAMES"]), int(
                entry_dict["EXIT FRAMES"]
            )
            entry_dict["frame_range"] = list(range(entry, exit + 1))
            col_name = entry_dict["ANIMAL"] + "_" + entry_dict["SHAPE"]
            self.data_df[col_name][
                self.data_df.index.isin(entry_dict["frame_range"])
            ] = 1

    def __calc_text_locs(self):
        add_spacer = 2
        self.loc_dict = {}
        for animal_cnt, animal_name in enumerate(
            self.roi_analyzer.multi_animal_id_list
        ):
            self.loc_dict[animal_name] = {}
            for shape in self.video_shapes:
                self.loc_dict[animal_name][shape] = {}
                self.loc_dict[animal_name][shape]["timer_text"] = "{} {} {}".format(
                    shape, animal_name, "timer:"
                )
                self.loc_dict[animal_name][shape]["entries_text"] = "{} {} {}".format(
                    shape, animal_name, "entries:"
                )
                self.loc_dict[animal_name][shape]["timer_text_loc"] = (
                    (self.video_meta_data["width"] + 5),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.scalers["space_size"] * add_spacer
                    ),
                )
                self.loc_dict[animal_name][shape]["timer_data_loc"] = (
                    int(self.border_img_w - (self.border_img_w / 8)),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.scalers["space_size"] * add_spacer
                    ),
                )
                add_spacer += 1
                self.loc_dict[animal_name][shape]["entries_text_loc"] = (
                    (self.video_meta_data["width"] + 5),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.scalers["space_size"] * add_spacer
                    ),
                )
                self.loc_dict[animal_name][shape]["entries_data_loc"] = (
                    int(self.border_img_w - (self.border_img_w / 8)),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.scalers["space_size"] * add_spacer
                    ),
                )
                add_spacer += 1

    def __create_counters(self):
        self.cnt_dict = {}
        for animal_cnt, animal_name in enumerate(
            self.roi_analyzer.multi_animal_id_list
        ):
            self.cnt_dict[animal_name] = {}
            for shape in self.video_shapes:
                self.cnt_dict[animal_name][shape] = {}
                self.cnt_dict[animal_name][shape]["timer"] = 0
                self.cnt_dict[animal_name][shape]["entries"] = 0
                self.cnt_dict[animal_name][shape]["entry_status"] = False

    def __calculate_cumulative(self):
        for animal in self.roi_analyzer.multi_animal_id_list:
            for shape in self.video_shapes:
                self.data_df["{}_{}_cum_sum_time".format(animal, shape)] = (
                    self.data_df["{}_{}".format(animal, shape)].cumsum() / self.fps
                )
                roi_bouts = list(
                    detect_bouts(
                        data_df=self.data_df,
                        target_lst=["{}_{}".format(animal, shape)],
                        fps=self.fps,
                    )["Start_frame"]
                )
                self.data_df["{}_{}_entry".format(animal, shape)] = 0
                self.data_df.loc[roi_bouts, "{}_{}_entry".format(animal, shape)] = 1
                self.data_df["{}_{}_cum_sum_entries".format(animal, shape)] = (
                    self.data_df["{}_{}_entry".format(animal, shape)].cumsum()
                )

    def __update_video_meta_data(self):
        new_cap = cv2.VideoCapture(self.video_path)
        new_cap.set(1, 1)
        _, img = self.cap.read()
        bordered_img = cv2.copyMakeBorder(
            img,
            0,
            0,
            0,
            int(self.video_meta_data["width"]),
            borderType=cv2.BORDER_CONSTANT,
            value=[0, 0, 0],
        )
        self.border_img_h, self.border_img_w = (
            bordered_img.shape[0],
            bordered_img.shape[1],
        )
        new_cap.release()

    def run(self):
        """
        Method to visualize ROI data. Results are stored in the `project_folder/frames/output/ROI_analysis`
        directory of the SimBA project.

        Returns
        -------
        None
        """

        video_timer = SimbaTimer(start=True)
        self.cap = cv2.VideoCapture(self.video_path)
        self.video_meta_data = get_video_meta_data(self.video_path)
        video_settings, pix_per_mm, self.fps = self.read_video_info(
            video_name=self.video_name
        )
        self.space_scale, radius_scale, res_scale, font_scale = 25, 10, 1500, 0.8
        max_dim = max(self.video_meta_data["width"], self.video_meta_data["height"])
        self.scalers = {}
        self.scalers["circle_size"], self.scalers["font_size"] = int(
            radius_scale / (res_scale / max_dim)
        ), float(font_scale / (res_scale / max_dim))
        self.scalers["space_size"] = int(self.space_scale / (res_scale / max_dim))
        color_lst = create_color_palettes(
            self.roi_analyzer.animal_cnt, int((len(self.roi_analyzer.bp_names) / 3))
        )[0]
        self.temp_folder = os.path.join(self.out_parent_dir, self.video_name, "temp")
        if not os.path.exists(self.temp_folder):
            os.makedirs(self.temp_folder)
        self.__update_video_meta_data()
        self.__calc_text_locs()
        self.__create_counters()
        self.__calculate_cumulative()

        data_arr = np.array_split(self.data_df, self.core_cnt)
        for df_cnt in range(len(data_arr)):
            data_arr[df_cnt]["group"] = df_cnt
        frm_per_core = len(data_arr[0])

        print(
            "Creating ROI images, multiprocessing (determined chunksize: {}, cores: {})...".format(
                str(self.multiprocess_chunksize), str(self.core_cnt)
            )
        )
        with multiprocessing.Pool(
            self.core_cnt, maxtasksperchild=self.maxtasksperchild
        ) as pool:
            constants = functools.partial(
                self.roi_plotter_mp,
                loc_dict=self.loc_dict,
                scalers=self.scalers,
                video_meta_data=self.video_meta_data,
                save_temp_directory=self.temp_folder,
                body_part_dict=self.bp_dict,
                input_video_path=self.video_path,
                roi_analyzer_data=self.roi_analyzer,
                video_shape_names=self.video_shapes,
                colors=color_lst,
                shape_meta_data=self.shape_dicts,
                style_attr=self.style_attr,
            )
            for cnt, result in enumerate(
                pool.imap(constants, data_arr, chunksize=self.multiprocess_chunksize)
            ):
                print(
                    "Image {}/{}, Video {}...".format(
                        str(int(frm_per_core * (result + 1))),
                        str(len(self.data_df)),
                        self.video_name,
                    )
                )
            print("Joining {} multi-processed video...".format(self.video_name))
            concatenate_videos_in_folder(
                in_folder=self.temp_folder,
                save_path=self.video_save_path,
                video_format="mp4",
            )

            video_timer.stop_timer()
            pool.terminate()
            pool.join()
            stdout_success(
                msg=f"Video {self.video_name} created. Video saved in project_folder/frames/output/ROI_analysis",
                elapsed_time=video_timer.elapsed_time_str,
            )


# test = ROIPlotMultiprocess(ini_path=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini',
#                video_path="termite_test.mp4",
#                core_cnt=5,
#                style_attr={'Show_body_part': True, 'Show_animal_name': True})
# test.run()

# test = ROIPlot(ini_path=r'/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini', video_path=r"Together_1.avi")
# test.insert_data()
# test.visualize_ROI_data()

# test = ROIPlot(ini_path=r"Z:\DeepLabCut\DLC_extract\Troubleshooting\ROI_2_animals\project_folder\project_config.ini", video_path=r"Z:\DeepLabCut\DLC_extract\Troubleshooting\ROI_2_animals\project_folder\videos\Video7.mp4")
# test.insert_data()
# test.visualize_ROI_data()

# test = ROIPlotMultiprocess(ini_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                video_path="Together_2.avi",
#                style_attr={'Show_body_part': True, 'Show_animal_name': False},
#                            core_cnt=5)
# test.run()



__author__ = "Simon Nilsson"

import logging

from simba.utils.enums import TagNames
from simba.utils.printing import log_event, stdout_warning


def log_warning(func):
    def wrapper(**kwargs):
        log_event(
            logger_name=f'{kwargs["source"] if "source" in kwargs.keys() else ""}.{func.__name__}',
            log_type=TagNames.WARNING.value,
            msg=kwargs["msg"],
        )
        results = func(**kwargs)
        stdout_warning(msg=f'{func.__name__}: {kwargs["msg"]}')
        return results

    return wrapper


def ThirdPartyAnnotationsOutsidePoseEstimationDataWarning(
    video_name: str,
    frm_cnt: int,
    log_status: bool = False,
    clf_name: str or None = None,
    annotation_frms: int or None = None,
    first_error_frm: int or None = None,
    ambiguous_cnt: int or None = None,
):
    if clf_name:
        msg = (
            f"SIMBA THIRD-PARTY ANNOTATION WARNING: SimBA found THIRD-PARTY annotations for behavior {clf_name} in video "
            f"{video_name} that are annotated to occur at times which is not present in the "
            f"video data you imported into SIMBA. The video you imported to SimBA has {str(frm_cnt)} frames. "
            f"However, in BORIS, you have annotated {clf_name} to happen at frame number {str(first_error_frm)}. "
            f"These ambiguous annotations occur in {str(ambiguous_cnt)} different frames for video {video_name} that SimBA will **remove** by default. "
            f"Please make sure you imported the same video as you annotated in BORIS into SimBA and the video is registered with the correct frame rate. "
            f"SimBA will only append annotations made to the frames present in the pose estimation data."
        )
    else:
        msg = f"SIMBA THIRD-PARTY ANNOTATION WARNING: The annotations for video {video_name} contain data for {str(annotation_frms)} frames. The pose-estimation features for the same video contain data for {str(frm_cnt)} frames. SimBA will use the annotations for the frames present in the pose-estimation data and discard the rest. If the annotation data is shorter than the pose-estimation data, SimBA will assume the missing annotation frames are all behavior absent."
    if log_status:
        logging.warning(msg=msg)
    stdout_warning(msg=msg)


def ThirdPartyAnnotationsClfMissingWarning(video_name: str, clf_name: str):
    msg = (
        f"SIMBA THIRD-PARTY ANNOTATION WARNING: No annotations detected for video {video_name} and behavior {clf_name}. "
        f"SimBA will set all frame annotations as absent."
    )
    stdout_warning(msg=msg)


def ThirdPartyAnnotationsAdditionalClfWarning(
    video_name: str, clf_names: list, source: str = "", log_status: bool = False
):
    msg = f"SIMBA THIRD-PARTY ANNOTATION WARNING: Annotations file for video {video_name} has annotations for the following behaviors {clf_names} that are NOT classifiers named in the SimBA project. SimBA will OMIT appending the data for these {str(len(clf_names))} classifiers."
    if log_status:
        logging.warning(msg=msg)
    stdout_warning(msg=msg)


def ThirdPartyAnnotationsInvalidFileFormatWarning(
    annotation_app: str, file_path: str, source: str = "", log_status: bool = False
):
    msg = f"SIMBA WARNING: {file_path} is not a valid {annotation_app} file and is skipped. See the SimBA GitHub repository for expected file format"
    if log_status:
        logging.warning(msg=msg)
    stdout_warning(msg=msg)


def ThirdPartyAnnotationsMissingAnnotationsWarning(
    video_name: str, clf_names: list, source: str = "", log_status: bool = False
):
    msg = f"SIMBA THIRD-PARTY ANNOTATION WARNING: No annotations detected for SimBA classifier(s) named {clf_names} for video {video_name}. All frame annotations will be set to behavior absent (0)."
    if log_status:
        logging.warning(msg=msg)
    stdout_warning(msg=msg)


def ThirdPartyAnnotationsFpsConflictWarning(
    video_name: str,
    annotation_fps: int,
    video_fps: int,
    source: str = "",
):
    msg = f"SIMBA THIRD-PARTY ANNOTATION WARNING: The FPS for video {video_name} is set to {str(video_fps)} in SimBA and {str(annotation_fps)} in the annotation file"
    stdout_warning(msg=msg)


def ThirdPartyAnnotationEventCountWarning(
    video_name: str,
    clf_name: str,
    start_event_cnt: int,
    stop_event_cnt: int,
    source: str = "",
    log_status: bool = False,
):
    msg = f"SIMBA THIRD-PARTY ANNOTATION WARNING: The annotations for behavior {clf_name} in video {video_name} contains {str(start_event_cnt)} start events and {str(stop_event_cnt)} stop events. SimBA requires the number of stop and start event counts to be equal. SimBA will try to find and delete the odd event stamps."
    if log_status:
        logging.warning(msg=msg)
    stdout_warning(msg=msg)


def ThirdPartyAnnotationOverlapWarning(
    video_name: str, clf_name: str, source: str = "", log_status: bool = False
):
    msg = f"SIMBA THIRD-PARTY ANNOTATION WARNING: The annotations for behavior {clf_name} in video {video_name} contains behavior start events that are initiated PRIOR to the PRECEDING behavior event ending. SimBA requires a specific behavior event to end before another behavior event can start. SimBA will try and delete these events."
    if log_status:
        logging.warning(msg=msg)
    stdout_warning(msg=msg)


def ThirdPartyAnnotationFileNotFoundWarning(
    video_name: str, source: str = "", log_status: bool = False
):
    msg = f"SIMBA THIRD-PARTY ANNOTATION WARNING: Could not find annotations for video features file {video_name} in the annotations directory."
    if log_status:
        logging.warning(msg=msg)
    stdout_warning(msg=msg)


@log_warning
def BodypartColumnNotFoundWarning(msg: str, source: str = ""):
    pass


@log_warning
def ShapWarning(msg: str, source: str = ""):
    pass


@log_warning
def InValidUserInputWarning(msg: str, source: str = ""):
    pass


@log_warning
def KleinbergWarning(msg: str, source: str = ""):
    pass


@log_warning
def NoFileFoundWarning(msg: str, source: str = ""):
    pass


@log_warning
def DataHeaderWarning(msg: str, source: str = ""):
    pass


@log_warning
def NoModuleWarning(msg: str, source: str = ""):
    pass


@log_warning
def FileExistWarning(msg: str, source: str = ""):
    pass


@log_warning
def DuplicateNamesWarning(msg: str, source: str = ""):
    pass


@log_warning
def InvalidValueWarning(msg: str, source: str = ""):
    pass


@log_warning
def NoDataFoundWarning(msg: str, source: str = ""):
    pass


@log_warning
def NotEnoughDataWarning(msg: str, source: str = ""):
    pass


@log_warning
def MissingUserInputWarning(msg: str, source: str = ""):
    pass


@log_warning
def SameInputAndOutputWarning(msg: str, source: str = ""):
    pass


@log_warning
def ROIWarning(msg: str, source: str = ""):
    pass


@log_warning
def MultiProcessingFailedWarning(msg: str):
    pass


@log_warning
def PythonVersionWarning(msg: str, source: str = ""):
    pass


@log_warning
def BorisPointEventsWarning(msg: str, source: str = ""):
    pass


@log_warning
def FFMpegCodecWarning(msg: str, source: str = ""):
    pass


@log_warning
def FFMpegNotFoundWarning(msg: str, source: str = ""):
    pass


@log_warning
def SkippingFileWarning(msg: str, source: str = ""):
    pass


@log_warning
def SkippingRuleWarning(msg: str, source: str = ""):
    pass


@log_warning
def IdenticalInputWarning(msg: str, source: str = ""):
    pass


@log_warning
def FrameRangeWarning(msg: str, source: str = ""):
    pass


@log_warning
def SamplingWarning(msg: str, source: str = ""):
    pass


@log_warning
def CropWarning(msg: str, source: str = ""):
    pass
